/*
 * @(#)TertrisN.c
 *
 * Copyright 1993 - 2008  David A. Bagley, bagleyd@tux.org
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995  Q. Alex Zhao, azhao@cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Next methods file for Tertris */

#include "file.h"
#include "rngs.h"
#include "TertrisNP.h"

static Point triangle_unit[MAX_ORIENT][3] = {
	{ {0, 0}, {-1, -1}, { 2, 0} },
	{ {0, 0}, { 1,  1}, {-2, 0} }
};
static Point triangle_list[MAX_ORIENT][4];
static Point triangle_border[MAX_ORIENT][4];

#ifdef WINVER
#ifndef LOGPATH
#define LOGPATH "/usr/tmp"
#endif
#ifndef INIFILE
#define INIFILE "wtertris.ini"
#endif

#define SECTION "setup"

#else

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif

static void resizeNextGame(NextTertrisWidget w);
static void destroyNextGame(Widget old);
#if 0
static void sizeNextGame(NextTertrisWidget w);
#endif
static Boolean setValuesNextGame(Widget current, Widget request, Widget renew);
static void quitNextGame(NextTertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void initializeNextGame(Widget request, Widget renew);
static void exposeNextGame(Widget renew, XEvent *event, Region region);
static void hideNextGame(NextTertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void enterNextGame(NextTertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void leaveNextGame(NextTertrisWidget w,
	XEvent *event, char **args, int nArgs);
static void showNextNextGame(NextTertrisWidget w,
	XEvent *event, char **args, int nArgs);

static char defaultTranslationsNextGame[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <KeyPress>n: ShowNext()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

static XtActionsRec actionsListNextGame[] =
{
	{(char *) "Quit", (XtActionProc) quitNextGame},
	{(char *) "Hide", (XtActionProc) hideNextGame},
	{(char *) "ShowNext", (XtActionProc) showNextNextGame},
	{(char *) "Enter", (XtActionProc) enterNextGame},
	{(char *) "Leave", (XtActionProc) leaveNextGame}
};

static XtResource resourcesNextGame[] =
{
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(NextTertrisWidget, core.width),
	 XtRString, (caddr_t) "132"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(NextTertrisWidget, core.height),
	 XtRString, (caddr_t) "136"},
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverseVideo, XtCReverseVideo, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(NextTertrisWidget, tertris.foreground),
	 XtRString, (caddr_t) XtDefaultForeground},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(NextTertrisWidget, tertris.background),
	 XtRString, (caddr_t) XtDefaultBackground},
	{XtNcolumns, XtCColumns, XtRInt, sizeof (int),
	 XtOffset(NextTertrisWidget, tertris.columns),
	 XtRString, (caddr_t) "10"},
	{XtNunits, XtCUnits, XtRInt, sizeof (int),
	 XtOffset(NextTertrisWidget, tertris.units),
	 XtRString, (caddr_t) "4"},
	{XtNdiagonal, XtCDiagonal, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.diagonal),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmixed, XtCMixed, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.mixed),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbonus, XtCBonus, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, nextTertris.newBonus),
	 XtRString, (caddr_t) "FALSE"},
	{XtNplainPiece, XtCPlainPiece, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.plainPiece),
	 XtRString, (caddr_t) "FALSE"},
	{XtNshowNext, XtCShowNext, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.showNext),
	 XtRString, (caddr_t) "TRUE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(NextTertrisWidget, tertris.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextTertrisWidget, tertris.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNrandomNumber, XtCRandomNumber, XtRInt, sizeof (int),
	 XtOffset(NextTertrisWidget, nextTertris.randomNumber),
	 XtRString, (caddr_t) "-1"},
	{XtNrandomRotation, XtCRandomRotation, XtRInt, sizeof (int),
	 XtOffset(NextTertrisWidget, nextTertris.randomRotation),
	 XtRString, (caddr_t) "-1"},
	{XtNrandomReflection, XtCRandomReflection, XtRInt, sizeof (int),
	 XtOffset(NextTertrisWidget, nextTertris.randomReflection),
	 XtRString, (caddr_t) "-1"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(NextTertrisWidget, tertris.select),
	 XtRCallback, (caddr_t) NULL}
};

NextTertrisClassRec nextTertrisClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "NextTertris",	/* class name */
		sizeof (NextTertrisRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) initializeNextGame,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListNextGame,	/* actions */
		XtNumber(actionsListNextGame),	/* num actions */
		resourcesNextGame,	/* resources */
		XtNumber(resourcesNextGame),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) destroyNextGame,	/* destroy */
		(XtWidgetProc) resizeNextGame,	/* resize */
		(XtExposeProc) exposeNextGame,	/* expose */
		(XtSetValuesFunc) setValuesNextGame,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsNextGame,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass nextTertrisWidgetClass = (WidgetClass) & nextTertrisClassRec;

static void
setNextGame(NextTertrisWidget w, int reason)
{
	tertrisCallbackStruct cb;

	cb.reason = reason;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}
#endif

#include "image.h"

static Pixmap dr = 0; /* dummy for future double buffering */

static void
checkThings(NextTertrisWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;

	if (w->tertris.units < MIN_UNITS) {
		w->tertris.units = MIN_UNITS;
		w->tertris.next.units = w->tertris.units;
		intCat(&buf1, "Number of units out of bounds, use at least ",
			MIN_UNITS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tertris.units);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	if (w->tertris.units > MAX_UNITS) {
		w->tertris.units = MAX_UNITS;
		w->tertris.next.units = w->tertris.units;
		intCat(&buf1, "Number of units out of bounds, use at most ",
			MAX_UNITS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->tertris.units);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
}

static void
draw3DFrame(const NextTertrisWidget w, Boolean inside,
	int x, int y, int width, int height)
{
	GC gc;

	if (inside)
		gc = w->tertris.inverseGC[3];
	else
		gc = w->tertris.inverseGC[1];
	FILLRECTANGLE(w, dr, gc,
		x, y, width, 1);
	FILLRECTANGLE(w, dr, gc,
		x, y + 1, 1, height - 1);
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + 1, width - 2, 1);
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + 2, 1, height - 3);
	if (inside)
		gc = w->tertris.inverseGC[0];
	else
		gc = w->tertris.inverseGC[1];
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + height - 1, width - 1, 1);
	FILLRECTANGLE(w, dr, gc,
		x + width - 1, y + 1, 1, height - 2);
	FILLRECTANGLE(w, dr, gc,
		x + 2, y + height - 2, width - 3, 1);
	FILLRECTANGLE(w, dr, gc,
		x + width - 2, y + 2, 1, height - 4);
}

static void
eraseFrame(NextTertrisWidget w)
{
	FILLRECTANGLE(w, dr, (w->tertris.showNext) ?
		w->tertris.inverseGC[2] : w->tertris.inverseGC[1],
		w->tertris.offset.x, w->tertris.offset.y,
		w->tertris.viewWidth, w->tertris.viewHeight);
}

static void
drawFrame(NextTertrisWidget w)
{
	int offsetX = w->tertris.offset.x - 2;
	int offsetY = w->tertris.offset.y - 2;
	int lengthX = w->tertris.offset.x + w->tertris.viewWidth + 2;
	int lengthY = w->tertris.offset.y + w->tertris.viewHeight + 2;

	if (w->tertris.showNext) {
		FILLRECTANGLE(w, dr, w->tertris.inverseGC[4],
			0, 0, w->core.width, offsetY);
		FILLRECTANGLE(w, dr, w->tertris.inverseGC[4],
			0, offsetY, offsetX, lengthY - offsetY);
		FILLRECTANGLE(w, dr, w->tertris.inverseGC[4],
			lengthX, offsetY,
			w->core.width - lengthX, lengthY - offsetY);
		FILLRECTANGLE(w, dr, w->tertris.inverseGC[4],
			0, lengthY, w->core.width, offsetY + 1);
		draw3DFrame(w, w->tertris.showNext,
			offsetX, offsetY,
			lengthX - offsetX, lengthY - offsetY);
	} else {
		FILLRECTANGLE(w, dr, w->tertris.inverseGC[1],
			w->tertris.offset.x, w->tertris.offset.y,
			w->tertris.viewWidth, w->tertris.viewHeight);
	}
}

static void
drawUnit(NextTertrisWidget w, int pmid, int cid, int i, int j)
{
	int orient = (i + j + w->tertris.nextThing.size) % MAX_ORIENT;

	triangle_list[orient][0].x = (i - 1) * (w->tertris.triWidth / 2) +
		w->tertris.offset.x;
	triangle_list[orient][0].y = (j - 2) * w->tertris.triHeight +
		 w->tertris.offset.y +
		((orient != 0) ? 0 : w->tertris.triHeight - 1);
	fillPolygon((TertrisWidget) w, pmid, cid, orient,
		triangle_list[orient], False);
}

#if 0
static void
clearThing(TertrisWidget w)
{
	XFillRectangle(display, mainWin, w->tertris.inverseGC[2],
		w->tertris.offset.x,
		topMidY - w->tertris.triHeight * HEX_ROW(HEX_SIZE) / 2,
		w->tertris.viewWidth,
		w->tertris.triHeight * HEX_ROW(HEX_SIZE) + 2);
}
#endif

static void
clearThing(NextTertrisWidget w)
{
	FILLRECTANGLE(w, dr, w->tertris.inverseGC[2],
		w->tertris.offset.x, w->tertris.offset.y,
		w->tertris.viewWidth, w->tertris.viewHeight);
}

static void
drawThing(NextTertrisWidget w)
{
	int i, j;
	int wid = HEX_COL(w->tertris.nextThing.size);
	int ht = HEX_ROW(w->tertris.nextThing.size);

	if (w->tertris.nextThing.polyNumber == IGNORE_RAND)
		return;
	clearThing(w);
	drawFrame(w);
	for (i = 0; i < wid; i++)
		for (j = 0; j < ht; j++)
			if (inHexagon(w->tertris.nextThing.size, i, j) != 0 &&
				tris[w->tertris.nextThing.units].polyiamond
				[w->tertris.nextThing.polyNumber].shape[j][i] != 0) {
				drawUnit(w, polyPixmap(i, j,
					(tris[w->tertris.nextThing.units].polyiamond
					[w->tertris.nextThing.polyNumber].shape),
					w->tertris.nextThing.size),
					w->tertris.nextThing.colorNumber,
					w->tertris.nextThing.xpos + i,
					w->tertris.nextThing.ypos + j);
			}
}

static void
redoNextNext(NextTertrisWidget w)
{
	redoNext((TertrisWidget) w);
	w->tertris.nextThing.xpos =
		(MAX_UNITS - w->tertris.nextThing.size) / 2;
	w->tertris.nextThing.ypos =
		(MAX_UNITS - w->tertris.nextThing.size) / 2;
#ifdef DEBUG
	(void) printf("Size: %d, Poly: %d, Color %d\n",
		w->tertris.nextThing.size, w->tertris.nextThing.polyNumber,
		w->tertris.nextThing.colorNumber);
#endif
}

#if 0
static void
resetThings(NextTertrisWidget w)
{
}

static void
resizeThings(NextTertrisWidget w)
{
}
#endif

#ifdef WINVER
static void
setValuesNextGame(NextTertrisWidget w)
{
	struct tagColor {
		int red, green, blue;
	} color;
	char szBuf[80], buf[20];
	int thing;

	w->tertris.units = GetPrivateProfileInt(SECTION, "units",
		DEFAULT_UNITS, INIFILE);
	w->tertris.diagonal = (BOOL) GetPrivateProfileInt(SECTION, "diagonal",
		DEFAULT_DIAGONAL, INIFILE);
	w->tertris.mixed = (BOOL) GetPrivateProfileInt(SECTION, "mixed",
		DEFAULT_MIXED, INIFILE);
	w->tertris.plainPiece = (BOOL) GetPrivateProfileInt(SECTION,
		"plainPiece", FALSE, INIFILE);
	w->tertris.showNext = (BOOL) GetPrivateProfileInt(SECTION, "showNext",
		FALSE, INIFILE);
	w->tertris.mono = (BOOL) GetPrivateProfileInt(SECTION, "mono",
		DEFAULT_MONO, INIFILE);
	w->tertris.reverse = (BOOL) GetPrivateProfileInt(SECTION,
		"reverseVideo", DEFAULT_REVERSE, INIFILE);
	/* #AEB2C3 */
	(void) GetPrivateProfileString(SECTION, "background", "255 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->tertris.inverseGC[1] = RGB(color.red, color.green, color.blue);
	w->tertris.inverseGC[0] = RGB(brighter(color.red),
		brighter(color.green), brighter(color.blue));
	w->tertris.inverseGC[2] = RGB(darker(color.red),
		darker(color.green), darker(color.blue));
	w->tertris.inverseGC[3] = RGB(darker(darker(color.red)),
		darker(darker(color.green)), darker(darker(color.blue)));
	w->tertris.inverseGC[4] = RGB(darker(darker(darker(color.red))),
		darker(darker(darker(color.green))),
		darker(darker(darker(color.blue))));
	for (thing = 0; thing < MAX_START_POLYIAMONDS; thing++) {
		(void) sprintf(buf, "thingColor%d", thing);
		(void) GetPrivateProfileString(SECTION, buf,
			polyiamondColorString[thing],
			szBuf, sizeof (szBuf), INIFILE);
		(void) sscanf(szBuf, "%d %d %d",
			&(color.red), &(color.green), &(color.blue));
		w->tertris.thingGC[thing][1] =
			RGB(color.red, color.green, color.blue);
		w->tertris.thingGC[thing][0] =
			RGB(brighter(color.red), brighter(color.green),
			brighter(color.blue));
		w->tertris.thingGC[thing][2] =
			RGB(darker(color.red), darker(color.green),
			darker(color.blue));
	}
}

void
destroyNextGame(HBRUSH brush)
{
	(void) DeleteObject(brush);
	/*PostQuitMessage(0);*/
}

#else
static void
getColor(NextTertrisWidget w, int thing)
{
	XGCValues values;
	XtGCMask valueMask;
	XColor colorCell, rgb;
	Boolean gotColor = False;

	valueMask = GCForeground | GCBackground;
	if (w->tertris.reverse) {
		values.background = w->tertris.foreground;
	} else {
		values.background = w->tertris.background;
	}
	if (!w->tertris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyiamondColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->tertris.thingColor[thing] =
				colorCell.pixel;
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyiamondColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyiamond ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->tertris.reverse) {
			values.background = w->tertris.foreground;
			values.foreground = w->tertris.background;
		} else {
			values.background = w->tertris.background;
			values.foreground = w->tertris.foreground;
		}
	}
	if (w->tertris.thingGC[thing][1])
		XtReleaseGC((Widget) w, w->tertris.thingGC[thing][1]);
	w->tertris.thingGC[thing][1] = XtGetGC((Widget) w, valueMask, &values);
	gotColor = False;
	if (!w->tertris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyiamondColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->tertris.thingColor[thing] =
				brighter((TertrisWidget) w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyiamondColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyiamond ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->tertris.reverse) {
			values.background = w->tertris.foreground;
			values.foreground = w->tertris.background;
		} else {
			values.background = w->tertris.background;
			values.foreground = w->tertris.foreground;
		}
	}
	if (w->tertris.thingGC[thing][0])
		XtReleaseGC((Widget) w, w->tertris.thingGC[thing][0]);
	w->tertris.thingGC[thing][0] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->tertris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyiamondColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->tertris.thingColor[thing] =
				darker((TertrisWidget) w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyiamondColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyiamond ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->tertris.reverse) {
			values.background = w->tertris.foreground;
			values.foreground = w->tertris.background;
		} else {
			values.background = w->tertris.background;
			values.foreground = w->tertris.foreground;
		}
	}
	if (w->tertris.thingGC[thing][2])
		XtReleaseGC((Widget) w, w->tertris.thingGC[thing][2]);
	w->tertris.thingGC[thing][2] = XtGetGC((Widget) w, valueMask,
		&values);
}

static void
setAllColors(NextTertrisWidget w)
{
	XGCValues values;
	XtGCMask valueMask;
	int polyiamond;

	valueMask = GCForeground | GCBackground;
	if (w->tertris.reverse) {
		values.background = w->tertris.background;
		values.foreground = w->tertris.foreground;
	} else {
		values.foreground = w->tertris.background;
		values.background = w->tertris.foreground;
	}
	if (w->tertris.inverseGC[1])
		XtReleaseGC((Widget) w, w->tertris.inverseGC[1]);
	w->tertris.inverseGC[1] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->tertris.mono) {
		values.foreground = brighter((TertrisWidget) w, (w->tertris.reverse) ?
			w->tertris.foreground : w->tertris.background);
	}
	if (w->tertris.inverseGC[0])
		XtReleaseGC((Widget) w, w->tertris.inverseGC[0]);
	w->tertris.inverseGC[0] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->tertris.mono) {
		values.foreground = darker((TertrisWidget) w, (w->tertris.reverse) ?
			w->tertris.foreground : w->tertris.background);
	}
	if (w->tertris.inverseGC[2])
		XtReleaseGC((Widget) w, w->tertris.inverseGC[2]);
	w->tertris.inverseGC[2] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->tertris.mono) {
		values.foreground = darker((TertrisWidget) w, darker((TertrisWidget) w, (w->tertris.reverse) ?
			w->tertris.foreground : w->tertris.background));
	}
	if (w->tertris.inverseGC[3])
		XtReleaseGC((Widget) w, w->tertris.inverseGC[3]);
	w->tertris.inverseGC[3] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->tertris.mono) {
		values.foreground = darker((TertrisWidget) w, darker((TertrisWidget) w, darker((TertrisWidget) w,
			(w->tertris.reverse) ? w->tertris.foreground :
			w->tertris.background)));
	}
	if (w->tertris.inverseGC[4])
		XtReleaseGC((Widget) w, w->tertris.inverseGC[4]);
	w->tertris.inverseGC[4] = XtGetGC((Widget) w, valueMask, &values);
	for (polyiamond = 0; polyiamond < MAX_START_POLYIAMONDS; polyiamond++)
		getColor(w, polyiamond);
}

static Boolean
setValuesNextGame(Widget current, Widget request, Widget renew)
{
	NextTertrisWidget c = (NextTertrisWidget) current;
	NextTertrisWidget w = (NextTertrisWidget) renew;
	Boolean redraw = False, setColors = False;

	checkThings(w);
	if (w->tertris.showNext != c->tertris.showNext) {
		redraw = True;
	}
	if (w->tertris.reverse != c->tertris.reverse ||
			w->tertris.mono != c->tertris.mono) {
		setAllColors(w);
		redraw = True;
	} else if (w->tertris.background != c->tertris.background ||
			w->tertris.foreground != c->tertris.foreground ||
			setColors) {
		setAllColors(w);
		redraw = True;
	} else if (w->tertris.plainPiece != c->tertris.plainPiece) {
		redraw = True;
	}
#if 0
	if (w->tertris.units != c->tertris.units) {
		w->tertris.next.units = w->tertris.units;
		resetThings(w);
	} else if (w->tertris.diagonal != c->tertris.diagonal) {
		w->tertris.next.diagonal = w->tertris.diagonal;
		resetThings(w);
	} else if (w->tertris.mixed != c->tertris.mixed) {
		w->tertris.next.mixed = w->tertris.mixed;
		resetThings(w);
	}
#endif
	if (w->tertris.menu != -1) {
		switch (w->tertris.menu) {
		case 3:
			w->tertris.menu = -1;
			clearThing(w);
			redraw = True;
			break;
		default:
			w->tertris.menu = -1;
			break;
		}
	}
	if (w->nextTertris.randomNumber == CLEARNEXT_RAND) {
		if (w->tertris.showNext) {
			drawThing(w);
		} else {
			clearThing(w);
		}
		w->nextTertris.randomNumber = IGNORE_RAND;
		redraw = True;
	} else if (w->nextTertris.randomNumber != IGNORE_RAND) {
		w->tertris.nextThing.randomNumber = w->nextTertris.randomNumber;
		w->tertris.nextThing.randomRotation = w->nextTertris.randomRotation;
		w->tertris.nextThing.randomReflection = w->nextTertris.randomReflection;
		w->tertris.nextThing.bonus = w->nextTertris.newBonus;
		w->tertris.next.units = w->tertris.units;
		w->tertris.next.diagonal = w->tertris.diagonal;
		w->tertris.next.mixed = w->tertris.mixed;
		redoNextNext(w);
		if (w->tertris.showNext) {
			drawThing(w);
		}
		w->nextTertris.randomNumber = IGNORE_RAND;
		redraw = True;
	}
	return (redraw);
}

static void
quitNextGame(NextTertrisWidget w, XEvent *event, char **args, int nArgs)
{
	Display *display = XtDisplay(w);

	XtCloseDisplay(display);
	exit(0);
}

static void
destroyNextGame(Widget old)
{
	NextTertrisWidget w = (NextTertrisWidget) old;
	int thing, i;

	for (thing = 0; thing < MAX_START_POLYIAMONDS; thing++) {
		for (i = 0; i < FG_SHADES; i++)
			XtReleaseGC(old, w->tertris.thingGC[thing][i]);
	}
	for (i = 0; i < BG_SHADES; i++)
		XtReleaseGC(old, w->tertris.inverseGC[i]);
	XtRemoveCallbacks(old, XtNselectCallback, w->tertris.select);
}
#endif

#define SQRT3_2 0.8660254 /* sqrt(3) / 2 */

#ifndef WINVER
static
#endif
void
resizeNextGame(NextTertrisWidget w)
{
	int i, j;
#ifdef WINVER
	RECT rect;

	/* Determine size of client area */
	(void) GetClientRect(w->core.hWnd, &rect);
	w->core.width = rect.right;
	w->core.height = rect.bottom;
#endif
	w->tertris.triHeight = MIN((w->core.height - 2) / (MAX_UNITS / 2 + 2),
		(w->core.width - 2) * SQRT3_2 / (MAX_UNITS / 2 + 1));
	w->tertris.triWidth = (int) (w->tertris.triHeight / SQRT3_2);
	w->tertris.viewWidth = w->tertris.triWidth *
		(MAX_UNITS / 2 + 1) + 1;
	w->tertris.viewHeight = w->tertris.triHeight * (MAX_UNITS / 2 + 2);
	w->tertris.offset.x = ((int) w->core.width -
		w->tertris.viewWidth) / 2;
	w->tertris.offset.y = ((int) w->core.height -
		w->tertris.viewHeight) / 2;

	for (j = 0; j < MAX_ORIENT; j++) {
		for (i = 0; i < 3; i++) {
			triangle_list[j][i].x =
				(w->tertris.triWidth - 1) / 2 *
				triangle_unit[j][i].x;
			triangle_list[j][i].y =
				(w->tertris.triHeight - 1) *
				triangle_unit[j][i].y;
			triangle_border[j][i].x =
				w->tertris.triWidth / 2 * triangle_unit[j][i].x;
				triangle_border[j][i].y =
				w->tertris.triHeight * triangle_unit[j][i].y;
		}
		triangle_list[j][3].x = triangle_list[j][1].x;
		triangle_list[j][3].y = -triangle_list[j][1].y;
		triangle_border[j][3].x = triangle_border[j][1].x;
		triangle_border[j][3].y = -triangle_border[j][1].y;
	}
}

#if 0
#ifndef WINVER
static
#endif
void
sizeNextGame(NextTertrisWidget w)
{
	resetThings(w);
	resizeNextGame(w);
}
#endif

#ifndef WINVER
static
#endif
void
initializeNextGame(
#ifdef WINVER
NextTertrisWidget w, HBRUSH brush
#else
Widget request, Widget renew
#endif
)
{
	int diagonal, mixed, turnStyle;
#ifndef WINVER
	int thing, i;
	NextTertrisWidget w = (NextTertrisWidget) renew;
#endif

#ifdef WINVER
	setValuesNextGame(w);
#else
	w->tertris.mono = (DefaultDepthOfScreen(XtScreen(w)) < 2 ||
		w->tertris.mono);
	for (thing = 0; thing < MAX_START_POLYIAMONDS; thing++) {
		for (i = 0; i < FG_SHADES; i++)
			 w->tertris.thingGC[thing][i] = NULL;
	}
	for (i = 0; i < BG_SHADES; i++)
		w->tertris.inverseGC[i] = NULL;
#endif

	w->tertris.quit = False;
	w->tertris.focus = False;
	w->tertris.next.units = w->tertris.units;
	w->tertris.next.diagonal = w->tertris.diagonal;
	w->tertris.next.mixed = w->tertris.mixed;
	w->tertris.nextThing.polyNumber = IGNORE_RAND;
	checkThings(w);
#if 0
	sizeNextGame(w);
#else
	resizeNextGame(w);
#endif
#ifdef WINVER
	brush = CreateSolidBrush(w->tertris.inverseGC[2]);
	SETBACK(w->core.hWnd, brush);
	(void) SRAND(time(NULL));
#else
	(void) SRAND(getpid());
	setAllColors(w);
#endif
	diagonal = (w->tertris.next.diagonal) ? 1 : 0;
	mixed = (w->tertris.next.mixed) ? 1 : 0;
	readThings();
	turnStyle = tris[w->tertris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	/* use intCat */
	if (turnStyle < 0 || turnStyle >= MAX_TURNSTYLE_TYPES) {
		(void) fprintf(stderr, "NextTertris: corrupted polyiamond file\n");
		(void) fprintf(stderr, "\tturnStyle = %d.\n", turnStyle);
		exit(1);
	}
	w->tertris.numberPolyiamonds = tris[w->tertris.next.units -
		MIN_UNITS].mode[diagonal].number[mixed][turnStyle];
	if (w->tertris.numberPolyiamonds <= 0 ||
			w->tertris.numberPolyiamonds > MAX_START_POLYIAMONDS) {
		(void) fprintf(stderr, "NextTertris: corrupted polyiamond file\n");
		(void) fprintf(stderr, "\tnumberPolyiamonds = %d.\n",
			w->tertris.numberPolyiamonds);
		exit(1);
	}
}

#ifndef WINVER
static
#endif
void
exposeNextGame(
#ifdef WINVER
NextTertrisWidget w
#else
Widget renew, XEvent *event, Region region
#endif
)
{
#ifndef WINVER
	NextTertrisWidget w = (NextTertrisWidget) renew;

	if (!w->core.visible)
		return;
#endif
	eraseFrame(w);
#if 0
	resizeThings(w);
#endif
	drawFrame(w);
	if (w->tertris.showNext) {
		drawThing(w);
	}
}

#ifndef WINVER
static
#endif
void
hideNextGame(NextTertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setNextGame(w, ACTION_HIDE);
}

#ifdef WINVER
void
valueNextGame(NextTertrisWidget w,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean diagonal, Boolean mixed, Boolean newBonus)
{
	w->tertris.nextThing.randomNumber = randomNumber;
	w->tertris.nextThing.randomRotation = randomRotation;
	w->tertris.nextThing.randomReflection = randomReflection;
	w->tertris.next.units = w->tertris.units = units;
	w->tertris.next.diagonal = w->tertris.diagonal = diagonal;
	w->tertris.next.mixed = w->tertris.mixed = mixed;
	w->tertris.nextThing.bonus = newBonus;
	redoNextNext(w);
	if (w->tertris.showNext) {
		drawThing(w);

	}
}
#endif

#ifndef WINVER
static
#endif
void
enterNextGame(NextTertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->tertris.focus = True;
	drawFrame(w);
}

#ifndef WINVER
static
#endif
void
leaveNextGame(NextTertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->tertris.focus = False;
	drawFrame(w);
}

#ifndef WINVER
static
#endif
void
showNextNextGame(NextTertrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->tertris.paused) {
		w->tertris.showNext = !w->tertris.showNext;
		if (w->tertris.showNext) {
			drawThing(w);
		} else {
			clearThing(w);
			drawFrame(w);
		}
		FLUSH(w);
	}
}
