/*
 * @(#)xwelltris.c
 *
 * Copyright 1993 - 2010  David A. Bagley, bagleyd@tux.org
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995  Q. Alex Zhao, azhao@cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Driver file for Welltris */

#ifndef WINVER
#include "version.h"
static const char aboutHelp[] = {
"Welltris Version " VERSION "\n"
"Send bugs (reports or fixes) to the author: "
"David Bagley <bagleyd@tux.org>\n"
"The latest version is at: "
"http://www.tux.org/~bagleyd/altris.html"
};

static const char optionsHelp[] = {
"[-geometry [{width}][x{height}][{+-}{xoff}[{+-}{yoff}]]]\n"
"[-display [{host}]:[{vs}]] [-[no]mono] [-[no]{reverse|rv}]\n"
"[-{foreground|fg} {color}] [-{background|bg} {color}]\n"
"[-[no]sound] [-bumpSound {filename}]\n"
"[-dripSound {filename}] [-level {starting_level}]\n"
"[-units {size_of_polyomino}] [-diameter {diameter}]\n"
"[-fill {rows}] [-[no]fixedFill] [-[no]suddenAppear]\n"
"[-[no]diagonal] [-[no]mixed] [-[no]bonus] [-[no]cw]\n"
"[-[no]plain] [-[no]showNext] [-[no]grid]\n"
"[-userName {string}] [-scoreFile {filename}]\n"
"[-scores] [-version]"
};
#endif

#if defined(HAVE_MOTIF) || defined(WINVER)
static const char descriptionHelp[] = {
"The ever popular falling tetromino game.  Enhanced to "
"allow all polyomino classes that do not have pieces with\n"
"holes.  Mixing polyomino classes (so as to permit polyominoes "
"made up by a variable number squares) can also\n"
"be selected, as well as, the 2 sets of diagonal polyominoes "
"that do not have pieces with holes.\n"
"\n"
"To make these games interesting and \"do-able\" the ability "
"to rotate gets turned off if it is easier than the 4\n"
"square polyomino.  An added ability \"reflect\" "
"(right-left -> left-right) allows one to attempt the 5 square and\n"
"6 square polyomino pieces.\n"
"\n"
"To get high scores, do not use \"show next\" mode, try "
"starting at a higher level and try to let \"welltris\" eat more\n"
"than one line at a time.  Polyominoes with diagonal or "
"with a higher number of squares earn more points.  Mixed\n"
"mode earns less points.  Play fast and see how well you "
"can do in the bonus mode.  Bonuses are set when the\n"
"level is changed automatically.  (Bonus mode not available "
"for 6 squares or 3 squares and diagonal mode).\n"
"\n"
"The program lists the top 15 players at the end, each player "
"has at most three scores."
};

static const char featuresHelp[] = {
"Click \"mouse-left\" button , or press \"J\", \"j\", "
"\"S\", or \"s\" keys to move the falling polyomino left.\n"
"\n"
"Click \"mouse-center\" button, or press \"K\", \"k\", "
"\"D\", or \"d\" keys to rotate the falling polyomino.\n"
"Note this may be turned off if the depending on the "
"\"turn-style\" but normally this is \"on\".\n"
"\n"
"Click \"mouse-right\" button, or press \"L\", \"l\", "
"\"F\", or \"f\" keys to move the falling polyomino right.\n"
"\n"
"Click \"mouse-scroll-up\", or press \"I\", \"i\", "
"\"E\", or \"e\" keys to flip (reflect) the falling polyomino.\n"
"Note this may be turned off if the depending on the "
"\"turn-style\" but normally this is \"off\".\n"
"\n"
"Click \"mouse-scroll-down\", or press \"space-bar\" "
"to drop the falling polyomino.\n"
"\n"
"The key pad and arrow keys are defined similar.\n"
"Key pad is defined for Welltris as:\n"
"  /    Reflect\n"
"  8    Rotate\n"
"  ^\n"
"4<5>6  Left, Rotate, Right\n"
"  v\n"
"  2    Drop\n"
"\n"
"Press \"C\" or \"c\" keys to clear.\n"
"\n"
"Press \">\" or \".\" keys to speed up "
"the falling polyomino.\n"
"\n"
"Press \"<\" or \",\" keys to slow down "
"the falling polyomino.\n"
"This may be turned off to avoid cheating.\n"
"\n"
"Press \"+\" or \"=\" keys to increase the "
"squares in the falling polyomino.  Note: To keep the\n"
"level of play similar, the \"turn style\" can change.  "
"The ability to reflect polyominoes is available\n"
"only when the number of squares >= 5.  The ability "
"to rotate polyominoes, is available only when\n"
"the number of squares is >= 4.  Bonus polyominoes "
"still have the current \"turn style\".\n"
"\n"
"Press \"-\" or \"_\" keys to decrease the squares in "
"the falling polyomino.\n"
"\n"
"Press \"]\" or \"}\" keys to increase the diameter/depth "
"size.\n"
"\n"
"Press \"[\" or \"{\" keys to decrease the diameter/depth "
"size.\n"
"\n"
"Press \"`\", \"1\", \"2\", \"3\", \"4\", \"5\", "
"\"6\", \"7\", \"8\", \"9\", or \"0\" keys\n"
"(not the keypad 0-9) to change the number of fill lines.  "
"\"`\" for 0 lines and \"0\" for 10 lines.\n"
"\n"
"Press \"X\" or \"x\" to toggle between fixed fill or "
"random fill.\n"
"\n"
"Press \"R\" or \"r\" to toggle between sudden initial "
"appearance or gradual initial appearance.\n"
"\n"
"Press \"G\" or \"g\" to toggle diagonal mode.  This "
"may not be available for current number of squares.\n"
"\n"
"Press \"M\" or \"m\" to toggle mixed mode.  This "
"may not be available for current number of squares.\n"
"\n"
"Press \"B\" or \"b\" to toggle bonus mode.  This "
"may not be available for current number of squares.\n"
"\n"
"Press \"W\" or \"w\" to toggle clockwise or counter "
"clockwise turning.\n"
"\n"
"Press \"P\" or \"p\" to toggle plain pieces or "
"three dimensional pieces.\n"
"\n"
"Press \"N\" or \"n\" to toggle show next.\n"
"\n"
"Press \"#\" key to toggle the grid.\n"
"\n"
"Press \"|\" key to pause.\n"
"\n"
"Press \"@\" key to toggle the sound.\n"
"\n"
"Press \"Esc\" key to hide and pause program.  When the "
"welltris window loses its focus, it also goes to\n"
"\"paused\" mode and it iconifies itself.\n"
"\n"
"Press \"Q\", \"q\", or \"CTRL-C\" keys to kill "
"program."
};

static const char referencesHelp[] = {
"Based at one time on Q. Alex Zhao's Tetris."
};
#endif

#include "file.h"
#ifdef WINVER
#include "WelltrisP.h"
#include "WelltrisNP.h"
#define TITLE "wwelltris"

static WelltrisRec widget;
static NextWelltrisRec widget2;

#ifndef SCOREPATH
#ifdef UNIXDELIM
#define SCOREPATH "c:/WINDOWS"
#else
#define SCOREPATH "c:\\WINDOWS"
#endif
#endif
#define PRINT_MESSAGE(b) (void) MessageBox(widget.core.hWnd, (LPCSTR) b, "Warning", MB_OK);
#else
#include "xwin.h"
#include <X11/Shell.h>
#include <X11/cursorfont.h>
#ifdef HAVE_MOTIF
#include <Xm/PanedW.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/MessageB.h>
#include <Xm/PushBG.h>
#include <Xm/CascadeB.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include <Xm/Form.h>
#include <Xm/DialogS.h>
#define PRINT_MESSAGE(b) printState(message, b)
#else
#define PRINT_MESSAGE(b) XtWarning(b)
#endif
#include "Welltris.h"
#include "WelltrisN.h"
#ifdef HAVE_XPM
#include <X11/xpm.h>
#ifdef CONSTPIXMAPS
#include "welltris.t.xpm"
#include "welltris.p.xpm"
#include "welltris.s.xpm"
#include "welltris.m.xpm"
#include "welltris.l.xpm"
#include "welltris.xpm"
#else
#include "pixmaps/welltris.t.xpm"
#include "pixmaps/welltris.p.xpm"
#include "pixmaps/welltris.s.xpm"
#include "pixmaps/welltris.m.xpm"
#include "pixmaps/welltris.l.xpm"
#include "pixmaps/welltris.xpm"
#endif
#define RESIZE_XPM(s) ((char **) (((s)<=32)?\
(((s)<=22)?(((s)<=16)?welltris_t_xpm:welltris_p_xpm):\
(((s)<=24)?welltris_s_xpm:welltris_m_xpm)):\
(((s)<=48)?welltris_l_xpm:welltris_xpm)))
#endif
#include "pixmaps/welltris.xbm"
#define DEFINE_XBM (char *) welltris_bits, welltris_width, welltris_height
#ifndef SCOREPATH
#ifdef VMS
#define SCOREPATH "SYS$LOGIN:"
#else
#define SCOREPATH "/var/games/altris"
#endif
#endif
#endif

#ifndef SCOREFILE
#define SCOREFILE "welltris.scores"
#endif

#define NEVER (-1)
#define FILE_NAME_LENGTH 1024
#define USER_NAME_LENGTH 120
#define DATELEN 27
#define MESSAGE_LENGTH (USER_NAME_LENGTH+64)
#define TITLE_LENGTH 2048
#define MAX_SCORES 15
#define MAX_USER_SCORES 3
#define NOACCESS ""
#define NOBODY "nobody"

typedef struct {
	int score, level, rows;
	char name[USER_NAME_LENGTH]; /* host[USER_NAME_LENGTH] */
	char date[DATELEN];
} GameRecord;

static GameRecord gameRecord[MAX_SCORES];
static int maxScore = 0;
static char messageDsp[MESSAGE_LENGTH] = "Welcome";

#ifdef HAVE_MOTIF
static Widget scoreText, maxScoreText, message, rowsText, levelSlider;
static Widget unitsSlider, fieldSlider, fillSlider;
static Widget fixedFillSwitch, suddenAppearSwitch;
static Widget diagonalSwitch, mixedSwitch, bonusSwitch;
static Widget cwSwitch, plainSwitch, showNextSwitch;
static Widget gridSwitch, soundSwitch;
static char buff[21];
static Widget setupDialog = NULL, descriptionDialog, featuresDialog;
static Widget optionsDialog, referencesDialog, aboutDialog;
static Widget restartDialog;
static Widget menuBar;
static Arg arg[3];
#else
#ifndef WINVER
static Widget shell;
#endif
static char titleDsp[TITLE_LENGTH] = "";
#endif
static char scoreFileName[FILE_NAME_LENGTH] = SCOREFILE;
#ifdef WINVER
#define PROGRAM_NAME_LENGTH 80
static char progDsp[PROGRAM_NAME_LENGTH] = TITLE;
static char userNameDsp[USER_NAME_LENGTH] = "Guest";
#else
static Pixmap pixmap = None;
static Widget topLevel, welltris, nextWelltris;
static char *progDsp;
static char userNameDsp[USER_NAME_LENGTH] = "";

#ifdef HAVE_MOTIF
static void
printState(Widget w, char *msg)
{
	XmString xmstr;

	if (!XtIsSubclass(w, xmLabelWidgetClass))
		XtError("printState() requires a Label Widget");
	xmstr = XmStringCreateLtoR(msg, XmSTRING_DEFAULT_CHARSET);
	XtVaSetValues(w, XmNlabelString, xmstr, NULL);
}
#endif

static void
printRecords(void)
{
	int i;

	(void) printf("	    WELLTRIS  HALL OF FAME\n\n");
	(void) printf("   # USER	     SCORE LEVEL  ROWS\n");
	for (i = 0; i < MAX_SCORES; i++) {
		if (gameRecord[i].score > 0) {
			(void) printf("%4d %-12s%9d %5d %5d\n", i + 1,
				gameRecord[i].name, gameRecord[i].score,
				gameRecord[i].level, gameRecord[i].rows);
		} else {
			break;
		}
	}
}
#endif

static void
initRecords(void)
{
	int i;

	for (i = 0; i < MAX_SCORES; i++) {
		gameRecord[i].score = NEVER;
		gameRecord[i].level = 0;
		gameRecord[i].rows = 0;
		(void) strncpy(gameRecord[i].name,
			NOACCESS, USER_NAME_LENGTH);
		gameRecord[i].date[0] = '\0';
	}
}

/* #define DEBUG 1 */

static void
readRecords(void)
{
	FILE *fp;
	int s = NEVER, l, r, i;
	char userName[USER_NAME_LENGTH];
	char *buf1 = NULL, *buf2 = NULL;
	char *fname, *lname, *name;

	stringCat(&buf1, CURRENTDELIM, scoreFileName);
	lname = buf1;
	stringCat(&buf1, SCOREPATH, FINALDELIM);
	stringCat(&buf2, buf1, SCOREFILE);
	free(buf1);
	fname = buf2;
	name = fname;
	if ((fp = fopen(name, "r")) == NULL) {
		/* Try current directory in case its not installed yet. */
		name = lname;
		if ((fp = fopen(name, "r")) == NULL) {
#if 0
			stringCat(&buf1, "Can not read ", fname);
			stringCat(&buf2, buf1, " or ");
			free(buf1);
			stringCat(&buf1, buf2, lname);
			free(buf2);
			PRINT_MESSAGE(buf1);
			free(buf1);
			free(lname);
			free(fname);
#endif
			return;
		}
/* Probably annoying */
#if 0
		else {
			stringCat(&buf1, "Can not read ", fname);
			stringCat(&buf2, buf1, ", falling back to ");
			free(buf1);
			stringCat(&buf1, buf2, lname);
			free(buf2);
			PRINT_MESSAGE(buf1);
			free(buf1);
		}
#endif
	}
	free(lname);
	free(fname);
	for (i = 0; i < MAX_SCORES; i++)
		gameRecord[i].score = 0;
	for (i = 0; i < MAX_SCORES; i++) {
		if (fscanf(fp, "%d %d %d %s\n", &s, &l, &r, userName) > 0) {
			gameRecord[i].score = s;
			gameRecord[i].level = l;
			gameRecord[i].rows = r;
			(void) strncpy(gameRecord[i].name,
				userName, USER_NAME_LENGTH);
			if (maxScore < gameRecord[i].score)
				maxScore = gameRecord[i].score;
		} else {
			break;
		}
	}
	(void) fclose(fp);
}

static void
writeRecords(void)
{
	FILE *fp;
	int i;
	char *buf1 = NULL, *buf2 = NULL;
	char *fname, *lname, *name;

	stringCat(&buf1, CURRENTDELIM, scoreFileName);
	lname = buf1;
	stringCat(&buf1, SCOREPATH, FINALDELIM);
	stringCat(&buf2, buf1, SCOREFILE);
	free(buf1);
	fname = buf2;
	name = fname;
	if ((fp = fopen(name, "w")) == NULL) {
		/* Try current directory in case its not installed yet. */
		name = lname;
		if ((fp = fopen(name, "w")) == NULL) {
			stringCat(&buf1, "Can not write to ", fname);
			stringCat(&buf2, buf1, " or ");
			free(buf1);
			stringCat(&buf1, buf2, lname);
			free(buf2);
			PRINT_MESSAGE(buf1);
			free(buf1);
			free(lname);
			free(fname);
			return;
		}
/* Probably annoying */
#if 0
		else {
			stringCat(&buf1, "Can not write to ", fname);
			stringCat(&buf2, buf1, ", falling back to ");
			free(buf1);
			stringCat(&buf1, buf2, lname);
			free(buf2);
			PRINT_MESSAGE(buf1);
			free(buf1);
		}
#endif
	}
	{
#if HAVE_FCNTL_H
		int lfd;
		char lockFile[FILE_NAME_LENGTH];

		(void) strncpy(lockFile, name, FILE_NAME_LENGTH - 6);
		lockFile[FILE_NAME_LENGTH - 6] = '\0';
		(void) strcat(lockFile, ".lock");
		while (((lfd = open(lockFile, O_CREAT | O_EXCL, 0644)) < 0) &&
				errno == EEXIST)
			(void) sleep(1);
		if (lfd < 0) {
#if 1
			(void) fprintf(stderr,
				"Lock file exists... guessing its an old one.\n");
#else
			(void) fprintf(stderr,
				"Lock file exists... score not recorded - sorry.\n");
			return;
#endif
		}
#endif
		for (i = 0; i < MAX_SCORES; i++) {
			if (gameRecord[i].score <= 0)
				break;
			(void) fprintf(fp, "%d %d %d %s\n",
				gameRecord[i].score,
				gameRecord[i].level,
				gameRecord[i].rows,
				gameRecord[i].name);
		}
		(void) fprintf(fp, "\n");
#if HAVE_FCNTL_H
		(void) close(lfd);
		(void) unlink(lockFile);
#endif
		(void) fclose(fp);
	}
	free(lname);
	free(fname);
}

#ifndef WINVER
/* There is probably a better way to assure that they are the same
 * but I do not know it off hand. */
static void
makeEquivalent(String *userName, String *scoreFile, int *level,
		int *units, Boolean *diagonal, Boolean *mixed, Boolean *bonus,
		Boolean *showNext)
{
	Boolean mono, reverse, plainPiece, grid;
	Boolean scoreOnly, versionOnly;
	Pixel foreground, background;

	XtVaGetValues(welltris,
		XtNuserName, userName,
		XtNscoreFile, scoreFile,
		XtNlevel, level,
		XtNunits, units,
		XtNdiagonal, diagonal,
		XtNmixed, mixed,
		XtNbonus, bonus,
		XtNplainPiece, &plainPiece,
		XtNshowNext, showNext,
		XtNgrid, &grid,
		XtNmono, &mono,
		XtNreverseVideo, &reverse,
		XtNforeground, &foreground,
		XtNbackground, &background,
		XtNscoreOnly, &scoreOnly,
		XtNversionOnly, &versionOnly, NULL);
	if (versionOnly) {
		(void) printf("%s\n", aboutHelp);
		exit(0);
	}
	if (scoreOnly) {
		initRecords();
		readRecords();
		printRecords();
		exit(0);
	}
	XtVaSetValues(nextWelltris,
		XtNlevel, *level,
		XtNunits, *units,
		XtNdiagonal, *diagonal,
		XtNmixed, *mixed,
		XtNbonus, *bonus,
		XtNplainPiece, plainPiece,
		XtNshowNext, *showNext,
		XtNgrid, grid,
		XtNmono, mono,
		XtNreverseVideo, reverse,
		XtNforeground, foreground,
		XtNbackground, background,
		NULL);
}
#endif

static void
printStatus(
#ifndef HAVE_MOTIF
	char *msg, int level,
#endif
	int numberRows, int yourScore)
{
	if (maxScore < yourScore) {
		maxScore = yourScore;
	}
#ifdef HAVE_MOTIF
	(void) sprintf(buff, "%d", numberRows);
	printState(rowsText, buff);
	(void) sprintf(buff, "%d", yourScore);
	printState(scoreText, buff);
	(void) sprintf(buff, "%d", maxScore);
	printState(maxScoreText, buff);
#else
	(void) sprintf(titleDsp, "%s: %d level %d rows @ (%d/%d) - %s",
		progDsp, level, numberRows, yourScore,
		maxScore, msg);
#ifdef WINVER
	SetWindowText(widget.core.hWnd, (LPSTR) titleDsp);
#else
	XtVaSetValues(XtParent(welltris), XtNtitle, titleDsp, NULL);
#endif
#endif
}

#define CPSCORE(a,b) a.score = b.score, a.level = b.level, a.rows = b.rows,\
	(void) strncpy(a.name, b.name, USER_NAME_LENGTH)

static Boolean
handleGameOver(int lastLevel, int numberRows, int yourScore)
{
	int i, j = 0, k = 0, rank = MAX_SCORES;
	GameRecord current, temp[MAX_SCORES];
	Boolean newRecord = False;

	readRecords();	/* Maybe its been updated by another */
	current.score = yourScore;
	current.level = lastLevel;
	current.rows = numberRows;
	(void) strncpy(current.name, userNameDsp, USER_NAME_LENGTH);
	for (i = 0; i < MAX_SCORES; i++) {
		if (j == 0 && current.score > gameRecord[i].score) {
			CPSCORE(temp[i], current);
			j = 1;
			rank = i;
		} else {
			CPSCORE(temp[i], gameRecord[i - j]);
		}
	}
	j = 0;
	if (rank < MAX_SCORES) {
		for (i = 0; i < MAX_SCORES; i++) {
			if (!strcmp(temp[i].name, userNameDsp)) {
				if (j < MAX_USER_SCORES) {
					CPSCORE(gameRecord[k], temp[i]);
					if (i == rank)
						newRecord = True;
					k++;
				}
				j++;
			} else {
				CPSCORE(gameRecord[k], temp[i]);
				k++;
			}
		}
		for (i = k; i < MAX_SCORES; i++) {
			gameRecord[i].score = 0;
		}
	}
	if (newRecord) {
		writeRecords();
		printStatus(
#ifndef HAVE_MOTIF
			messageDsp, lastLevel,
#endif
			numberRows, yourScore);
		return True;
	}
	return False;
}

#ifdef WINVER
static LRESULT CALLBACK
showNextCB(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	HBRUSH brush = (HBRUSH) NULL;
	PAINTSTRUCT paint;

	if (widget2.core.hWnd == NULL)
		widget2.core.hWnd = hDlg;
	if (GetFocus()) {
		if (!widget2.welltris.focus) {
			widget2.core.hDC = GetDC(widget2.core.hWnd);
			(void) SelectObject(widget2.core.hDC,
				GetStockObject(NULL_BRUSH));
			enterNextGame(&widget2);
			(void) EndPaint(widget2.core.hWnd, &paint);
		}
	} else {
		if (widget2.welltris.focus) {
			widget2.core.hDC = GetDC(widget2.core.hWnd);
			(void) SelectObject(widget2.core.hDC,
				GetStockObject(NULL_BRUSH));
			leaveNextGame(&widget2);
			(void) EndPaint(widget2.core.hWnd, &paint);
		}
	}
	switch (message) {
	case WM_DESTROY:
		destroyNextGame(brush);
		(void) DestroyWindow(widget2.core.hWnd);
		widget2.core.hWnd = NULL;
		return TRUE;
	case WM_SIZE:
		resizeNextGame(&widget2);
		(void) InvalidateRect(widget2.core.hWnd, NULL, TRUE);
		break;
	case WM_PAINT:
		exposeNextGame(&widget2);
		(void) InvalidateRect(widget2.core.hWnd, NULL, TRUE);
		break;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case ACTION_EXIT:
			destroyNextGame(brush);
			(void) DestroyWindow(widget2.core.hWnd);
			widget2.core.hWnd = NULL;
			return TRUE;
		case ACTION_HIDE:
			hideNextGame(&widget2);
			break;
#if 0
		case ACTION_CLEAR:
			widget2.core.hDC = GetDC(widget2.core.hWnd);
			(void) SelectObject(widget2.core.hDC,
				GetStockObject(NULL_PEN));
			showNextGame(&widget2);
			(void) ReleaseDC(widget2.core.hWnd, widget2.core.hDC);
			break;
#endif
		case ACTION_SHOW_NEXT:
			if (widget2.welltris.showNext) {
				ShowWindow(widget2.core.hWnd,
					SW_SHOWNOACTIVATE);
			} else {
				ShowWindow(widget2.core.hWnd, SW_HIDE);
			}
			break;
#if 0
		case ACTION_SHOW_NEXT:
			if (widget2.core.hWnd == NULL)
				widget2.core.hWnd =
					CreateDialog(widget2.core.hInstance,
					"ShowNext", widget2.core.hWnd,
					(DLGPROC) showNextCB);
			else {
				destroyNextGame(brush);
				(void) DestroyWindow(widget2.core.hWnd);
				widget2.core.hWnd = NULL;
				return TRUE;
			}
			break;
#endif
		}
		break;
	/* default:
		return (DefWindowProc(widget2.core.hWnd, message,
			wParam, lParam)); */
	}
	return FALSE;
}
#endif

static void
initialize(
#ifdef WINVER
WelltrisWidget w, HBRUSH brush, HBRUSH brush2
#else
Widget w
#endif
)
{
	int level, units;
	Boolean diagonal, mixed, bonus, showNext;
	char *userName, *scoreFile;

#ifdef WINVER
	initializeGame(w, brush);
	level = w->welltris.level;
	units = widget2.welltris.units = w->welltris.units;
	diagonal = widget2.welltris.diagonal = w->welltris.diagonal;
	mixed = widget2.welltris.mixed = w->welltris.mixed;
	bonus = widget2.welltris.bonus = w->welltris.bonus;
	userName = w->welltris.userName;
	scoreFile = w->welltris.scoreFile;
	initializeNextGame(&widget2, brush2);
	(void) CreateDialog(widget.core.hInstance,
		"ShowNext", widget.core.hWnd, (DLGPROC) showNextCB);
	if (!widget2.welltris.showNext)
		ShowWindow(widget2.core.hWnd, SW_HIDE);
	showNext = widget.welltris.showNext = widget2.welltris.showNext;
#else
	makeEquivalent(&userName, &scoreFile, &level, &units,
		&diagonal, &mixed, &bonus, &showNext);
#ifdef HAVE_MOTIF
	XmScaleSetValue(levelSlider, level);
#endif
#endif
	initRecords();
	readRecords();
#ifndef WINVER
	(void) strncpy(userNameDsp, userName, USER_NAME_LENGTH);
#endif
	if (!strcmp(userName, "(null)") || !strcmp(userName, NOBODY) ||
			!strcmp(userName, NOACCESS)) {
#ifdef WINVER
		(void) strncpy(userNameDsp, userName, USER_NAME_LENGTH);
#else
		char * login = getlogin();

		if (login == NULL) {
			(void) strcpy(userNameDsp, "");
		} else {
			(void) sprintf(userNameDsp, "%s", login);
		}
		if (!strcmp(userNameDsp, "(null)") ||
				!strcmp(userNameDsp, NOBODY) ||
				!strcmp(userNameDsp, NOACCESS))
			/* It really IS nobody */
			(void) sprintf(userNameDsp, "%s", "guest");
#endif
	}
	if (strcmp(scoreFile, ""))
		(void) strncpy(scoreFileName, scoreFile, FILE_NAME_LENGTH);
	printStatus(
#ifndef HAVE_MOTIF
		messageDsp, level,
#endif
		0, 0);
}

#ifdef WINVER
void
setGame(WelltrisWidget w, int reason)
#else
static void
callbackGame(Widget w, caddr_t clientData, welltrisCallbackStruct * callData)
#endif
{
#ifndef WINVER
	int reason = callData->reason;
#endif
	int level, units, diameter, fill, score, completedRows;
	Boolean fixedFill, suddenAppear, diagonal, mixed, bonus;
	Boolean cw, plainPiece, grid, showNext, sound;

	(void) strcpy(messageDsp, "");
#ifdef WINVER
	level = w->welltris.level;
	units = w->welltris.units;
	diameter = w->welltris.diameter;
	fill = w->welltris.fill;
	fixedFill = w->welltris.fixedFill;
	suddenAppear = w->welltris.suddenAppear;
	diagonal = w->welltris.next.diagonal;
	mixed = w->welltris.next.mixed;
	bonus = w->welltris.bonus;
	cw = w->welltris.cw;
	plainPiece = w->welltris.plainPiece;
	grid = w->welltris.grid;
	showNext = w->welltris.showNext;
	sound = w->welltris.sound;
	completedRows = w->welltris.completedRows;
	score = w->welltris.score;
#else
	XtVaGetValues(w,
		XtNlevel, &level,
		XtNunits, &units,
		XtNdiameter, &diameter,
		XtNfill, &fill,
		XtNfixedFill, &fixedFill,
		XtNsuddenAppear, &suddenAppear,
		XtNdiagonal, &diagonal,
		XtNmixed, &mixed,
		XtNbonus, &bonus,
		XtNcw, &cw,
		XtNplainPiece, &plainPiece,
		XtNgrid, &grid,
		XtNsound, &sound,
		XtNcompletedRows, &completedRows,
		XtNscore, &score, NULL);
	XtVaGetValues(nextWelltris,
		XtNshowNext, &showNext, NULL);
#endif
	switch (reason) {
	case ACTION_HIDE:
#ifdef WINVER
		ShowWindow(w->core.hWnd, SW_SHOWMINIMIZED);
#else
		(void) XIconifyWindow(XtDisplay(topLevel),
			XtWindow(topLevel),
			XScreenNumberOfScreen(XtScreen(topLevel)));
#ifndef HAVE_MOTIF
		(void) XIconifyWindow(XtDisplay(shell),
			XtWindow(shell),
			XScreenNumberOfScreen(XtScreen(shell)));
#endif
#endif
		break;
#ifndef WINVER
	case ACTION_RESTART_QUERY:
#ifdef HAVE_MOTIF
		XtManageChild(restartDialog);
#else
		XtVaSetValues(welltris,
			XtNmenu, 4, NULL); /* menu choice */
#endif
		break;
#endif
	case ACTION_STATUS:
		printStatus(
#ifndef HAVE_MOTIF
			messageDsp, level,
#endif
			completedRows, score);
#ifdef HAVE_MOTIF
		XmScaleSetValue(levelSlider, level);
#endif
		break;
#ifndef WINVER
	case ACTION_DRAW_NEXT:
		XtVaSetValues(nextWelltris,
			XtNrandomNumber, callData->randomNumber,
			XtNrandomRotation, callData->randomRotation,
			XtNrandomReflection, callData->randomReflection,
			XtNunits, callData->units,
			XtNdiagonal, callData->diagonal,
			XtNmixed, callData->mixed,
			XtNbonus, callData->newBonus, NULL);
		break;
#endif
	case ACTION_SHOW_NEXT:
		showNext = !showNext;
#ifdef WINVER
		showNextNextGame(&widget2);
#else
		XtVaSetValues(nextWelltris,
			XtNshowNext, showNext, NULL);
#ifdef HAVE_MOTIF
		XtVaSetValues(nextWelltris,
			XtNrandomNumber, CLEARNEXT_RAND, NULL);
		if (setupDialog != NULL)
			XmToggleButtonSetState(showNextSwitch,
				showNext, True);
#else
		if (showNext) {
			int x, y, width;

			XtVaGetValues(topLevel,
				XtNx, &x,
				XtNy, &y,
				XtNwidth, &width, NULL);
			XtVaSetValues(shell,
				XtNx, x + width,
				XtNy, y, NULL);
			XtRealizeWidget(shell);
			XtVaSetValues(nextWelltris,
				XtNrandomNumber, CLEARNEXT_RAND,
				NULL);
		} else {
			XtVaSetValues(nextWelltris,
				XtNrandomNumber, CLEARNEXT_RAND,
				NULL);
			XtUnrealizeWidget(shell);
		}
#endif
#endif
		break;
	case ACTION_DEC_UNITS:
		units--;
#ifdef WINVER
		w->welltris.units = units;
#else
		XtVaSetValues(welltris,
			XtNunits, units, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmScaleSetValue(unitsSlider, units);
#endif
#endif
		break;
	case ACTION_INC_UNITS:
		units++;
#ifdef WINVER
		w->welltris.units = units;
#else
		XtVaSetValues(welltris,
			XtNunits, units, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmScaleSetValue(unitsSlider, units);
#endif
#endif
		break;
	case ACTION_DEC_DIAMETER:
		diameter--;
#ifdef WINVER
		w->welltris.diameter = diameter;
#else
		XtVaSetValues(welltris,
			XtNdiameter, diameter, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmScaleSetValue(fieldSlider, diameter);
#endif
#endif
		break;
	case ACTION_INC_DIAMETER:
		diameter++;
#ifdef WINVER
		w->welltris.diameter = diameter;
#else
		XtVaSetValues(welltris,
			XtNdiameter, diameter, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmScaleSetValue(fieldSlider, diameter);
#endif
#endif
		break;
	case ACTION_FIXED_FILL:
		fixedFill = !fixedFill;
#ifdef WINVER
		w->welltris.fixedFill = fixedFill;
#else
		XtVaSetValues(welltris,
			XtNfixedFill, fixedFill, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(fixedFillSwitch,
				fixedFill, True);
#endif
#endif
		break;
	case ACTION_SUDDEN_APPEAR:
		suddenAppear = !suddenAppear;
#ifdef WINVER
		w->welltris.suddenAppear = suddenAppear;
#else
		XtVaSetValues(welltris,
			XtNsuddenAppear, suddenAppear, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(suddenAppearSwitch,
				suddenAppear, True);
#endif
#endif
		break;
	case ACTION_DIAGONAL:
		diagonal = !diagonal;
#ifdef WINVER
		w->welltris.diagonal = diagonal;
#else
		XtVaSetValues(welltris,
			XtNdiagonal, diagonal, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(diagonalSwitch,
				diagonal, True);
#endif
#endif
		break;
	case ACTION_MIXED:
		mixed = !mixed;
#ifdef WINVER
		w->welltris.mixed = mixed;
#else
		XtVaSetValues(welltris,
			XtNmixed, mixed, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(mixedSwitch,
				mixed, True);
#endif
#endif
		break;
	case ACTION_BONUS:
		bonus = !bonus;
#ifdef WINVER
		w->welltris.bonus = bonus;
#else
		XtVaSetValues(welltris,
			XtNbonus, bonus, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(bonusSwitch,
				bonus, True);
#endif
#endif
		break;
	case ACTION_CW:
		cw = !cw;
#ifdef WINVER
		w->welltris.cw = cw;
#else
		XtVaSetValues(welltris,
			XtNcw, cw, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(cwSwitch, cw, True);
#endif
#endif
		break;
	case ACTION_PLAIN_PIECE:
		plainPiece = !plainPiece;
#ifdef WINVER
		w->welltris.plainPiece = plainPiece;
		widget2.welltris.plainPiece = plainPiece;
#else
		XtVaSetValues(welltris,
			XtNplainPiece, plainPiece, NULL);
		XtVaSetValues(nextWelltris,
			XtNplainPiece, plainPiece, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(plainSwitch, plainPiece,
				True);
#endif
#endif
		break;
	case ACTION_GRID:
		grid = !grid;
#ifdef WINVER
		w->welltris.grid = grid;
		widget2.welltris.grid = grid;
#else
		XtVaSetValues(welltris,
			XtNgrid, grid, NULL);
		XtVaSetValues(nextWelltris,
			XtNgrid, grid, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(gridSwitch, grid, True);
#endif
#endif
		break;
	case ACTION_SOUND:
		sound = !sound;
#ifdef WINVER
		w->welltris.sound = sound;
#else
		XtVaSetValues(welltris,
			XtNsound, sound, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmToggleButtonSetState(soundSwitch,
				sound, True);
#endif
#endif
		break;
	case ACTION_FILL0:
	case ACTION_FILL1:
	case ACTION_FILL2:
	case ACTION_FILL3:
	case ACTION_FILL4:
	case ACTION_FILL5:
	case ACTION_FILL6:
	case ACTION_FILL7:
	case ACTION_FILL8:
	case ACTION_FILL9:
	case ACTION_FILL10:
		fill = reason - ACTION_FILL0;
#ifdef WINVER
		w->welltris.fill = fill;
#else
		XtVaSetValues(welltris,
			XtNfill, fill, NULL);
#ifdef HAVE_MOTIF
		if (setupDialog != NULL)
			XmScaleSetValue(fillSlider, fill);
#endif
#endif
		break;
	case ACTION_GAME_OVER:
		if (handleGameOver(level, completedRows, score))
			(void) sprintf(messageDsp,
				"Congratulations %s!!", userNameDsp);
		else
			(void) strncpy(messageDsp, "Game Over!", MESSAGE_LENGTH);
		break;
	}
#ifdef HAVE_MOTIF
	printState(message, messageDsp);
#endif
}

#ifdef WINVER
void
setGameNext(WelltrisWidget w, int reason,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean diagonal, Boolean mixed, Boolean newBonus)
{
	if (reason == ACTION_DRAW_NEXT) {
		valueNextGame(&widget2,
			randomNumber, randomRotation, randomReflection,
			units, diagonal, mixed, newBonus);
	}
}
#endif

#ifdef WINVER
void
setNextGame(NextWelltrisWidget w, int reason)
#else
static void
callbackNextGame(Widget w, caddr_t clientData, welltrisCallbackStruct * callData)
#endif
{
#ifndef WINVER
	int reason = callData->reason;
#endif

	switch (reason) {
	case ACTION_HIDE:
#ifdef WINVER
		ShowWindow(w->core.hWnd, SW_SHOWMINIMIZED);
#else
		(void) XIconifyWindow(XtDisplay(topLevel),
			XtWindow(topLevel),
			XScreenNumberOfScreen(XtScreen(topLevel)));
#ifndef HAVE_MOTIF
		(void) XIconifyWindow(XtDisplay(shell),
			XtWindow(shell),
			XScreenNumberOfScreen(XtScreen(shell)));
#endif
#endif
		break;
	}
}

#ifdef WINVER
static LRESULT CALLBACK
about(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	if (message == WM_COMMAND && LOWORD(wParam) == IDOK) {
		(void) EndDialog(hDlg, TRUE);
		return TRUE;
	}
	return FALSE;
}

static LRESULT CALLBACK
WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	HBRUSH brush = (HBRUSH) NULL;
	HBRUSH brush2 = (HBRUSH) NULL;
	PAINTSTRUCT paint;

	widget.core.hWnd = hWnd;
	if (GetFocus()) {
		if (!widget.welltris.focus) {
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_BRUSH));
			enterGame(&widget);
			(void) EndPaint(hWnd, &paint);
		}
	} else {
		if (widget.welltris.focus) {
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_BRUSH));
			leaveGame(&widget);
			(void) EndPaint(hWnd, &paint);
		}
	}
	switch (message) {
	case WM_CREATE:
		initialize(&widget, brush, brush2);
		break;
	case WM_DESTROY:
		destroyGame(brush);
		break;
	case WM_SIZE:
		resizeGame(&widget);
		(void) InvalidateRect(hWnd, NULL, TRUE);
		break;
	case WM_PAINT:
		widget.core.hDC = BeginPaint(hWnd, &paint);
		(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
		exposeGame(&widget);
		(void) EndPaint(hWnd, &paint);
		break;
	case WM_TIMER:
		KillTimer(hWnd, 1);
		widget.core.hDC = GetDC(hWnd);
		(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
		falldownGame(&widget);
		(void) ReleaseDC(hWnd, widget.core.hDC);
		SetTimer(hWnd, 1,
			(unsigned) 7 * widget.welltris.delay, NULL);
		break;
	case WM_RBUTTONDOWN:
		widget.core.hDC = GetDC(hWnd);
		(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
		(void) moveGameInput(&widget, MOVE_RIGHT);
		(void) ReleaseDC(hWnd, widget.core.hDC);
		break;
	case WM_MBUTTONDOWN:
		widget.core.hDC = GetDC(hWnd);
		(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
		(void) moveGameInput(&widget, ROTATE);
		(void) ReleaseDC(hWnd, widget.core.hDC);
		break;
	case WM_LBUTTONDOWN:
		widget.core.hDC = GetDC(hWnd);
		(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
		(void) moveGameInput(&widget, MOVE_LEFT);
		(void) ReleaseDC(hWnd, widget.core.hDC);
		break;
#if (_WIN32_WINNT >= 0x0400) || (_WIN32_WINDOWS > 0x0400)
	case WM_MOUSEWHEEL:
		widget.core.hDC = GetDC(hWnd);
		(void) SelectObject(widget.core.hDC, GetStockObject(NULL_PEN));
		{
			int zDelta = ((short) HIWORD(wParam));

			if (zDelta > (WHEEL_DELTA >> 1)) {
				(void) moveGameInput(&widget, REFLECT);
			} else if (zDelta < -(WHEEL_DELTA >> 1)) {
				(void) moveGameInput(&widget, DROP);
			}
		}
		(void) ReleaseDC(hWnd, widget.core.hDC);
		break;
#endif
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case ACTION_EXIT:
			destroyGame(brush);
			break;
		case ACTION_HIDE:
			hideGame(&widget);
			break;
		case ACTION_CLEAR:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_PEN));
			clearGame(&widget);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
		case ACTION_RIGHT:
		case ACTION_LEFT:
		case ACTION_ROTATE:
		case ACTION_REFLECT:
		case ACTION_DROP:
			{
				widget.core.hDC = GetDC(hWnd);
				(void) SelectObject(widget.core.hDC,
					GetStockObject(NULL_PEN));
				(void) moveGameInput(&widget,
					(int) LOWORD(wParam) - ACTION_DROP + 1);
				(void) ReleaseDC(hWnd, widget.core.hDC);
			}
			break;
		case ACTION_INC_LEVEL:
			(void) incrementGameLevel(&widget);
			break;
		case ACTION_DEC_LEVEL:
			(void) decrementGameLevel(&widget);
			break;
		case ACTION_INC_UNITS:
			(void) incrementGameUnits(&widget);
			break;
		case ACTION_DEC_UNITS:
			(void) decrementGameUnits(&widget);
			break;
		case ACTION_INC_DIAMETER:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_PEN));
			(void) incrementGameDiameter(&widget);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
		case ACTION_DEC_DIAMETER:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_PEN));
			(void) decrementGameDiameter(&widget);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
		case ACTION_FIXED_FILL:
			(void) fixedfillGame(&widget);
			break;
		case ACTION_SUDDEN_APPEAR:
			(void) suddenAppearGame(&widget);
			break;
		case ACTION_DIAGONAL:
			(void) diagonalGame(&widget);
			break;
		case ACTION_MIXED:
			(void) mixedGame(&widget);
			break;
		case ACTION_BONUS:
			(void) bonusGame(&widget);
			break;
		case ACTION_CW:
			(void) cwGame(&widget);
			break;
		case ACTION_PLAIN_PIECE:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_PEN));
			(void) plainPieceGame(&widget);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
		case ACTION_GRID:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_PEN));
			(void) gridGame(&widget);
			(void) ReleaseDC(hWnd, widget.core.hDC);
			break;
		case ACTION_SOUND:
			(void) soundGame(&widget);
			break;
		case ACTION_FILL0:
		case ACTION_FILL1:
		case ACTION_FILL2:
		case ACTION_FILL3:
		case ACTION_FILL4:
		case ACTION_FILL5:
		case ACTION_FILL6:
		case ACTION_FILL7:
		case ACTION_FILL8:
		case ACTION_FILL9:
		case ACTION_FILL10:
			widget.core.hDC = GetDC(hWnd);
			(void) SelectObject(widget.core.hDC,
				GetStockObject(NULL_PEN));
			(void) fillGame(&widget,
				(int) LOWORD(wParam) - ACTION_FILL0);
				(void) ReleaseDC(hWnd, widget.core.hDC);
				break;
		case ACTION_SHOW_NEXT:
			showNextNextGame(&widget2);
			if (widget2.welltris.showNext) {
				ShowWindow(widget2.core.hWnd,
					SW_SHOWNOACTIVATE);
			} else {
				ShowWindow(widget2.core.hWnd, SW_HIDE);
			}
#if 0
			if (widget2.core.hWnd == NULL) {
				widget2.core.hWnd =
					CreateDialog(widget.core.hInstance,
					"ShowNext", hWnd, (DLGPROC) showNextCB);
			} else {
				destroyNextGame(brush2);
				(void) DestroyWindow(widget2.core.hWnd);
				widget2.core.hWnd = NULL;
			}
#endif
			break;
		case ACTION_ABOUT:
			(void) DialogBox(widget.core.hInstance,
				"About", hWnd, (DLGPROC) about);
			break;
		case ACTION_DESCRIPTION:
			(void) MessageBox(hWnd, descriptionHelp,
				"Description", MB_OK);
			break;
		case ACTION_FEATURES:
			(void) MessageBox(hWnd, featuresHelp,
				"Features", MB_OK);
			break;
		case ACTION_REFERENCES:
			(void) MessageBox(hWnd, referencesHelp,
				"References", MB_OK);
			break;
		}
		break;
	default:
		return (DefWindowProc(hWnd, message, wParam, lParam));
	}
	return FALSE;
}

int WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine,
		int numCmdShow)
{
	HWND hWnd;
	MSG msg;
	WNDCLASS wc;
	HACCEL hAccel;

	if (!hPrevInstance) {
		wc.style = CS_HREDRAW | CS_VREDRAW;
		wc.lpfnWndProc = WindowProc;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = 0;
		wc.hInstance = hInstance;
		wc.hIcon = LoadIcon(hInstance, TITLE);
		wc.hCursor = LoadCursor((HINSTANCE) NULL, IDC_ARROW);
		wc.hbrBackground = (HBRUSH) GetStockObject(GRAY_BRUSH);
		wc.lpszMenuName = TITLE;
		wc.lpszClassName = TITLE;
		if (!RegisterClass(&wc))
			return FALSE;
	}
	widget.core.hInstance = hInstance;
	hWnd = CreateWindow(TITLE,
		TITLE,
		WS_OVERLAPPEDWINDOW,
		(signed) CW_USEDEFAULT,
		(signed) CW_USEDEFAULT,
		446,
		484,
		HWND_DESKTOP,
		(HMENU) NULL,
		hInstance,
		(void *) NULL);
	if (!hWnd)
		return FALSE;
	hAccel = (HACCEL) LoadAccelerators(hInstance, TITLE);
	(void) ShowWindow(hWnd, numCmdShow);
	SetTimer(hWnd, 1, 1000, NULL);
	(void) UpdateWindow(hWnd);
	while (GetMessage(&msg, (HWND) NULL, 0, 0))
		if (!IsDialogMessage(widget2.core.hWnd, &msg)) {
			if (!TranslateAccelerator(hWnd, hAccel, &msg)) {
				(void) TranslateMessage(&msg);
				(void) DispatchMessage(&msg);
			}
		}
	KillTimer(hWnd, 1);
	return (msg.wParam);
}

#else

static void
usage(char * programName)
{
	unsigned int i;

	(void) fprintf(stderr, "usage: %s\n", programName);
	for (i = 0; i < strlen(optionsHelp); i++) {
		if (i == 0 || optionsHelp[i - 1] == '\n')
			(void) fprintf(stderr, "\t");
		(void) fprintf(stderr, "%c", (optionsHelp[i]));
	}
	(void) fprintf(stderr, "\n");
	exit(1);
}

static XrmOptionDescRec options[] =
{
	{(char *) "-mono", (char *) "*welltris.mono", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nomono", (char *) "*welltris.mono", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-rv", (char *) "*welltris.reverseVideo", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-reverse", (char *) "*welltris.reverseVideo", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-norv", (char *) "*welltris.reverseVideo", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-noreverse", (char *) "*welltris.reverseVideo", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-fg", (char *) "*welltris.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-foreground", (char *) "*welltris.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-bg", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-background", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-level", (char *) "*welltris.level", XrmoptionSepArg, NULL},
	{(char *) "-units", (char *) "*welltris.units", XrmoptionSepArg, NULL},
	{(char *) "-diameter", (char *) "*welltris.diameter", XrmoptionSepArg, NULL},
	{(char *) "-fill", (char *) "*welltris.fill", XrmoptionSepArg, NULL},
	{(char *) "-fixedFill", (char *) "*welltris.fixedFill", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nofixedFill", (char *) "*welltris.fixedFill", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-suddenAppear", (char *) "*welltris.suddenAppear", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nosuddenAppear", (char *) "*welltris.suddenAppear", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-diagonal", (char *) "*welltris.diagonal", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nodiagonal", (char *) "*welltris.diagonal", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-mixed", (char *) "*welltris.mixed", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nomixed", (char *) "*welltris.mixed", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-bonus", (char *) "*welltris.bonus", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nobonus", (char *) "*welltris.bonus", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-cw", (char *) "*welltris.cw", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nocw", (char *) "*welltris.cw", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-plain", (char *) "*welltris.plain", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-noplain", (char *) "*welltris.plain", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-showNext", (char *) "*welltris.showNext", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-noshowNext", (char *) "*welltris.showNext", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-grid", (char *) "*welltris.grid", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nogrid", (char *) "*welltris.grid", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-sound", (char *) "*welltris.sound", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nosound", (char *) "*welltris.sound", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-bumpSound", (char *) "*welltris.bumpSound", XrmoptionSepArg, NULL},
	{(char *) "-dripSound", (char *) "*welltris.dripSound", XrmoptionSepArg, NULL},
	{(char *) "-userName", (char *) "*welltris.userName", XrmoptionSepArg, NULL},
	{(char *) "-scoreFile", (char *) "*welltris.scoreFile", XrmoptionSepArg, NULL},
	{(char *) "-scores", (char *) "*welltris.scoreOnly", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-version", (char *) "*welltris.versionOnly", XrmoptionNoArg, (char *) "TRUE"}
};

static void
playingTimer(XtPointer clientData, XtIntervalId *id)
{
	int delay;

	XtVaGetValues(welltris,
		XtNdelay, &delay, NULL);
	XtVaSetValues(welltris,
		XtNact, True, NULL);
	(void) XtAddTimeOut(7 * delay, playingTimer,
		clientData);
}

#ifdef HAVE_MOTIF
static void
callbackGameRestart(Widget w, XtPointer clientData, XmAnyCallbackStruct * cbs)
{
	if (cbs->reason == XmCR_OK) {
		XtVaSetValues(welltris, XtNmenu, 4, NULL); /* menu choice */
	}
}

static void
levelChangeListener(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int level = cbs->value, old;

	XtVaGetValues(welltris,
		XtNlevel, &old, NULL);
	if (old != level) {
		XtVaSetValues(welltris,
			XtNlevel, level, NULL);
	}
}

static void
unitsChangeListener(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int units = cbs->value, old;
	Boolean diagonal, mixed, bonus;

	XtVaGetValues(welltris,
		XtNunits, &old,
		XtNdiagonal, &diagonal,
		XtNmixed, &mixed,
		XtNbonus, &bonus, NULL);
	if (old == units) {
		return;
	}
	if (LEGAL(units, diagonal, mixed, bonus)) {
		XtVaSetValues(welltris,
			XtNunits, units, NULL);
	} else if (LEGAL(units, diagonal, mixed, !bonus)) {
		/* Try toggling bonus */
		XmToggleButtonSetState(bonusSwitch, !bonus, True);
		XtVaSetValues(welltris,
			XtNunits, units,
			XtNbonus, !bonus, NULL);
	} else if (LEGAL(units, diagonal, !mixed, bonus)) {
		/* Try toggling mixed */
		XmToggleButtonSetState(mixedSwitch, !mixed, True);
		XtVaSetValues(welltris,
			XtNunits, units,
			XtNmixed, !mixed, NULL);
	} else if (LEGAL(units, !diagonal, mixed, bonus)) {
		/* Try toggling diagonal */
		XmToggleButtonSetState(diagonalSwitch, !diagonal, True);
		XtVaSetValues(welltris,
			XtNunits, units,
			XtNdiagonal, !diagonal, NULL);
	} else {
		/* Give up */
		XmScaleSetValue(unitsSlider, old);
	}
}

static void
fieldChangeListener(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int diameter = cbs->value, old;

	XtVaGetValues(welltris,
		XtNdiameter, &old, NULL);
	if (old != diameter) {
		XtVaSetValues(welltris,
			XtNdiameter, diameter, NULL);
	}
}
static void
fillChangeListener(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int fill = cbs->value, old;

	XtVaGetValues(welltris,
		XtNfill, &old, NULL);
	if (old != fill) {
		XtVaSetValues(welltris,
			XtNfill, fill, NULL);
	}
}

static void
fixedFillToggle(Widget w, XtPointer clientData,
		XmToggleButtonCallbackStruct * cbs)
{
	Boolean fixedFill = cbs->set;

	XtVaSetValues(welltris,
		XtNfixedFill, fixedFill, NULL);
}

static void
suddenAppearToggle(Widget w, XtPointer clientData,
		XmToggleButtonCallbackStruct * cbs)
{
	Boolean suddenAppear = cbs->set;

	XtVaSetValues(welltris,
		XtNsuddenAppear, suddenAppear, NULL);
}

static void
diagonalToggle(Widget w, XtPointer clientData,
		XmToggleButtonCallbackStruct * cbs)
{
	int units;
	Boolean diagonal = cbs->set, mixed, bonus;

	XtVaGetValues(welltris,
		XtNunits, &units,
		XtNmixed, &mixed,
		XtNbonus, &bonus, NULL);
	if (LEGAL(units, diagonal, mixed, bonus)) {
		XtVaSetValues(welltris,
			XtNdiagonal, diagonal, NULL);
	} else if (LEGAL(units, diagonal, mixed, !bonus)) {
		/* Try toggling bonus */
		XmToggleButtonSetState(bonusSwitch, !bonus, True);
		XtVaSetValues(welltris,
			XtNdiagonal, diagonal,
			XtNbonus, !bonus, NULL);
	} else if (LEGAL(units, diagonal, !mixed, bonus)) {
		/* Try toggling mixed */
		XmToggleButtonSetState(mixedSwitch, !mixed, True);
		XtVaSetValues(welltris,
			XtNdiagonal, diagonal,
			XtNmixed, !mixed, NULL);
	} else {
		/* Give up */
		XmToggleButtonSetState(diagonalSwitch, !diagonal, True);
	}
}

static void
mixedToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	int units;
	Boolean mixed = cbs->set, diagonal, bonus;

	XtVaGetValues(welltris,
		XtNunits, &units,
		XtNdiagonal, &diagonal,
		XtNbonus, &bonus, NULL);
	if (LEGAL(units, diagonal, mixed, bonus)) {
		XtVaSetValues(welltris,
			XtNmixed, mixed, NULL);
	} else if (LEGAL(units, diagonal, mixed, !bonus)) {
		/* Try toggling bonus */
		XmToggleButtonSetState(bonusSwitch, !bonus, True);
		XtVaSetValues(welltris,
			XtNmixed, mixed,
			XtNbonus, !bonus, NULL);
	} else {
		/* Give up */
		XmToggleButtonSetState(mixedSwitch, !mixed, True);
	}
}

static void
bonusToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	int units;
	Boolean bonus = cbs->set, diagonal, mixed;

	XtVaGetValues(welltris,
		XtNunits, &units,
		XtNdiagonal, &diagonal,
		XtNmixed, &mixed,
		NULL);
	if (LEGAL(units, diagonal, mixed, bonus)) {
		XtVaSetValues(welltris,
			XtNbonus, bonus, NULL);
	} else {
		/* Give up */
		XmToggleButtonSetState(bonusSwitch, !bonus, True);
	}
}

static void
cwToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	Boolean cw = cbs->set;

	XtVaSetValues(welltris,
		XtNcw, cw, NULL);
}

static void
plainToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	Boolean plainPiece = cbs->set;

	XtVaSetValues(welltris,
		XtNplainPiece, plainPiece, NULL);
	XtVaSetValues(nextWelltris,
		XtNplainPiece, plainPiece, NULL);
}

static void
showNextToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	Boolean showNext = cbs->set;

	XtVaSetValues(nextWelltris,
		XtNshowNext, showNext, NULL);
}

static void
gridToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	Boolean grid = cbs->set;

	XtVaSetValues(welltris,
		XtNgrid, grid, NULL);
	XtVaSetValues(nextWelltris,
		XtNgrid, grid, NULL);
}

static void
soundToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	Boolean sound = cbs->set;

	XtVaSetValues(welltris,
		XtNsound, sound, NULL);
}

static void
fileMenuListener(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 0)
		exit(0);
}

static void
closeDialog(Widget w, Widget shellDialog)
{
	XtDestroyWidget(shellDialog);
	setupDialog = NULL;
}

static Widget
createSetup(Widget w, char *title)
{
	Widget button, dialog, pane;
	Widget sliderRowCol, switchRowCol;
	int units, diameter, fill;
	Boolean fixedFill, suddenAppear, diagonal, mixed;
	Boolean bonus, cw, plainPiece, showNext, grid, sound;
	char titleDsp[FILE_NAME_LENGTH + 8];
	XmString titleString = NULL;

	(void) sprintf(titleDsp, "%s: %s\n", progDsp, title);
	titleString = XmStringCreateSimple((char *) titleDsp);
	XtSetArg(arg[0], XmNdeleteResponse, XmDESTROY);
	XtSetArg(arg[1], XmNdialogTitle, titleString);
	dialog = XmCreateDialogShell(w, (char *) "Setup", arg, 2);
	XmStringFree(titleString);
	pane = XtVaCreateWidget("pane", xmPanedWindowWidgetClass, dialog,
		XmNsashWidth, 1,
		XmNsashHeight, 1, NULL);
	sliderRowCol = XtVaCreateManagedWidget("sliderRowCol",
		xmRowColumnWidgetClass, pane,
		XmNnumColumns, 3,
		XmNorientation, XmHORIZONTAL,
		XmNpacking, XmPACK_COLUMN, NULL);
	switchRowCol = XtVaCreateManagedWidget("switchRowCol",
		xmRowColumnWidgetClass, pane,
		XmNnumColumns, 10,
		XmNorientation, XmHORIZONTAL,
		XmNpacking, XmPACK_COLUMN, NULL);
	XtVaGetValues(welltris,
		XtNunits, &units,
		XtNdiameter, &diameter,
		XtNfill, &fill,
		XtNfixedFill, &fixedFill,
		XtNsuddenAppear, &suddenAppear,
		XtNdiagonal, &diagonal,
		XtNmixed, &mixed,
		XtNbonus, &bonus,
		XtNcw, &cw,
		XtNplainPiece, &plainPiece,
		XtNshowNext, &showNext,
		XtNgrid, &grid,
		XtNsound, &sound, NULL);
	unitsSlider = XtVaCreateManagedWidget("unitsSlider",
		xmScaleWidgetClass, sliderRowCol,
		XtVaTypedArg, XmNtitleString, XmRString, "Polyomino Size", 15,
		XmNminimum, MIN_UNITS,
		XmNmaximum, MAX_UNITS,
		XmNvalue, units,
		XmNshowValue, True,
		XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(unitsSlider, XmNvalueChangedCallback,
		(XtCallbackProc) unitsChangeListener, (XtPointer) NULL);
	fieldSlider = XtVaCreateManagedWidget("fieldSlider",
		xmScaleWidgetClass, sliderRowCol,
		XtVaTypedArg, XmNtitleString, XmRString, "Field Size", 11,
		XmNminimum, MIN_DIAMETER,
		XmNmaximum, MAX_DIAMETER,
		XmNvalue, diameter,
		XmNshowValue, True,
		XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(fieldSlider, XmNvalueChangedCallback,
		(XtCallbackProc) fieldChangeListener, (XtPointer) NULL);
	fillSlider = XtVaCreateManagedWidget("fillSlider",
		xmScaleWidgetClass, sliderRowCol,
		XtVaTypedArg, XmNtitleString, XmRString, "Fill", 5,
		XmNminimum, 0,
		XmNmaximum, MAX_FILL,
		XmNvalue, fill,
		XmNshowValue, True,
		XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(fillSlider, XmNvalueChangedCallback,
		(XtCallbackProc) fillChangeListener, (XtPointer) NULL);
	fixedFillSwitch = XtVaCreateManagedWidget("Fixed Fill",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(fixedFillSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) fixedFillToggle, (XtPointer) NULL);
	suddenAppearSwitch = XtVaCreateManagedWidget("Sudden Appear",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(suddenAppearSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) suddenAppearToggle, (XtPointer) NULL);
	diagonalSwitch = XtVaCreateManagedWidget("Diagonal",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(diagonalSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) diagonalToggle, (XtPointer) NULL);
	mixedSwitch = XtVaCreateManagedWidget("Mixed",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(mixedSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) mixedToggle, (XtPointer) NULL);
	bonusSwitch = XtVaCreateManagedWidget("Bonus",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(bonusSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) bonusToggle, (XtPointer) NULL);
	cwSwitch = XtVaCreateManagedWidget("Clockwise",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(cwSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) cwToggle, (XtPointer) NULL);
	plainSwitch = XtVaCreateManagedWidget("Plain Piece",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(plainSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) plainToggle, (XtPointer) NULL);
	showNextSwitch = XtVaCreateManagedWidget("Show Next",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(showNextSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) showNextToggle, (XtPointer) NULL);
	gridSwitch = XtVaCreateManagedWidget("Grid",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(gridSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) gridToggle, (XtPointer) NULL);
	soundSwitch = XtVaCreateManagedWidget("Sound",
		xmToggleButtonWidgetClass, switchRowCol, NULL);
	XtAddCallback(soundSwitch, XmNvalueChangedCallback,
		(XtCallbackProc) soundToggle, (XtPointer) NULL);
	XmToggleButtonSetState(fixedFillSwitch, fixedFill, True);
	XmToggleButtonSetState(suddenAppearSwitch, suddenAppear, True);
	XmToggleButtonSetState(diagonalSwitch, diagonal, True);
	XmToggleButtonSetState(mixedSwitch, mixed, True);
	XmToggleButtonSetState(bonusSwitch, bonus, True);
	XmToggleButtonSetState(cwSwitch, cw, True);
	XmToggleButtonSetState(plainSwitch, plainPiece, True);
	XmToggleButtonSetState(showNextSwitch, showNext, True);
	XmToggleButtonSetState(gridSwitch, grid, True);
	XmToggleButtonSetState(soundSwitch, sound, True);
	button = XtVaCreateManagedWidget("OK",
		xmPushButtonGadgetClass, pane,
		XmNtopAttachment, XmATTACH_FORM,
		XmNbottomAttachment, XmATTACH_FORM,
		XmNleftAttachment, XmALIGNMENT_CENTER,
		XmNshowAsDefault, True,
		XmNdefaultButtonShadowThickness, 1,
		NULL);
	XtAddCallback(button, XmNactivateCallback,
		(XtCallbackProc) closeDialog, dialog);
	XtManageChild(pane);
	return dialog;
}

static void
playCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 2) {
		/*XtManageChild(setupDialog); */
		if (setupDialog == NULL)
			setupDialog = createSetup(menuBar, (char *) "Setup");
		return;
	}
	XtVaSetValues(welltris,
		XtNmenu, val + 1, NULL); /* Q */
}

static Widget
createQuery(Widget w, char *text, char *title, XtCallbackProc callback)
{
	Widget button, messageBox;
	char titleDsp[FILE_NAME_LENGTH + 8];
	XmString titleString = NULL, messageString = NULL;
	static XmStringCharSet charSet =
		(XmStringCharSet) XmSTRING_DEFAULT_CHARSET;

	messageString = XmStringCreateLtoR(text, charSet);
	(void) sprintf(titleDsp, "%s: %s", progDsp, title);
	titleString = XmStringCreateSimple((char *) titleDsp);
	XtSetArg(arg[0], XmNdialogTitle, titleString);
	XtSetArg(arg[1], XmNmessageString, messageString);
	messageBox = XmCreateWarningDialog(w, (char *) "queryBox",
		arg, 2);
	button = XmMessageBoxGetChild(messageBox, XmDIALOG_HELP_BUTTON);
	XtUnmanageChild(button);
	XmStringFree(titleString);
	XmStringFree(messageString);
	XtAddCallback(messageBox, XmNokCallback, callback, (XtPointer) NULL);
	XtAddCallback(messageBox, XmNcancelCallback, callback,
		(XtPointer) NULL);
	return messageBox;
}

static Widget
createHelp(Widget w, char *text, char *title)
{
	Widget button, messageBox;
	char titleDsp[FILE_NAME_LENGTH + 8];
	XmString titleString = NULL, messageString = NULL, buttonString = NULL;
	static XmStringCharSet charSet =
		(XmStringCharSet) XmSTRING_DEFAULT_CHARSET;

	messageString = XmStringCreateLtoR(text, charSet);
	(void) sprintf(titleDsp, "%s: %s", progDsp, title);
	titleString = XmStringCreateSimple((char *) titleDsp);
	buttonString = XmStringCreateSimple((char *) "OK");
	XtSetArg(arg[0], XmNdialogTitle, titleString);
	XtSetArg(arg[1], XmNokLabelString, buttonString);
	XtSetArg(arg[2], XmNmessageString, messageString);
	messageBox = XmCreateInformationDialog(w, (char *) "helpBox",
		arg, 3);
	button = XmMessageBoxGetChild(messageBox, XmDIALOG_CANCEL_BUTTON);
	XtUnmanageChild(button);
	button = XmMessageBoxGetChild(messageBox, XmDIALOG_HELP_BUTTON);
	XtUnmanageChild(button);
	XmStringFree(titleString);
	XmStringFree(buttonString);
	XmStringFree(messageString);
	return messageBox;
}

static void
helpMenuListener(Widget w, XtPointer value, XtPointer clientData)
{
	int val = (int) value;

	switch (val) {
	case 0:
		XtManageChild(descriptionDialog);
		break;
	case 1:
		XtManageChild(featuresDialog);
		break;
	case 2:
		XtManageChild(optionsDialog);
		break;
	case 3:
		XtManageChild(referencesDialog);
		break;
	case 4:
		XtManageChild(aboutDialog);
		break;
	default:
		{
			char *buf;

			intCat(&buf, "helpMenuListener: %d\n", val);
			XtWarning(buf);
			free(buf);
		}
	}
}
#endif

int
main(int argc, char **argv)
{
	int pixmapSize;
#ifdef HAVE_MOTIF
	Widget pullDownMenu, widget;
	Widget menuBarPanel, mainPanel, controlPanel;
	Widget scoreRowCol, levelRowCol;
	Widget messageRowCol;
	XmString fileString, playString;
	XmString quitString;
	XmString pauseString, clearString, setupString;
#if 0
	XmString incrementString, decrementString;
	XmString fixedFillString, suddenAppearString;
	XmString diagonalString, mixedString, bonusString;
	XmString cwString, plainString, showNextString;
#endif
#else
	Boolean showNext;
	int x, y, width;
#endif

	progDsp = argv[0];
	topLevel = XtInitialize(argv[0], "Welltris",
		options, XtNumber(options), &argc, argv);
	if (argc != 1)
		usage(argv[0]);

#ifdef HAVE_MOTIF
	menuBarPanel = XtVaCreateManagedWidget("menuBarPanel",
		xmPanedWindowWidgetClass, topLevel,
		XmNseparatorOn, False,
		XmNsashWidth, 1,
		XmNsashHeight, 1, NULL);
	fileString = XmStringCreateSimple((char *) "File");
	playString = XmStringCreateSimple((char *) "Play");
	menuBar = XmVaCreateSimpleMenuBar(menuBarPanel, (char *) "menuBar",
		XmVaCASCADEBUTTON, fileString, 'F',
		XmVaCASCADEBUTTON, playString, 'P',
		NULL);
	XmStringFree(fileString);
	XmStringFree(playString);
	quitString = XmStringCreateSimple((char *) "Exit");
	(void) XmVaCreateSimplePulldownMenu(menuBar, (char *) "file_menu",
		0, fileMenuListener,
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, quitString, 'x', NULL, NULL,
		NULL);
	XmStringFree(quitString);
	pauseString = XmStringCreateSimple((char *) "Pause");
	clearString = XmStringCreateSimple((char *) "Clear");
	setupString = XmStringCreateSimple((char *) "Setup");
#if 0
	incrementString = XmStringCreateSimple((char *) "Increment");
	decrementString = XmStringCreateSimple((char *) "Decrement");
	fixedFillString = XmStringCreateSimple((char *) "Fixed Fill");
	suddenAppearString = XmStringCreateSimple((char *) "Sudden Appear");
	diagonalString = XmStringCreateSimple((char *) "Diagonal");
	mixedString = XmStringCreateSimple((char *) "Mixed");
	bonusString = XmStringCreateSimple((char *) "Bonus");
	cwString = XmStringCreateSimple((char *) "Cw");
	plainString = XmStringCreateSimple((char *) "Plain Piece");
	showNextString = XmStringCreateSimple((char *) "Show Next");
#endif
	(void) XmVaCreateSimplePulldownMenu(menuBar, (char *) "play_menu",
		1, playCB,
		XmVaPUSHBUTTON, pauseString, 'P', NULL, NULL,
		XmVaPUSHBUTTON, clearString, 'C', NULL, NULL,
		XmVaPUSHBUTTON, setupString, 'S', NULL, NULL,
#if 0
		XmVaPUSHBUTTON, incrementString, 'I', NULL, NULL,
		XmVaPUSHBUTTON, decrementString, 'D', NULL, NULL,
		XmVaPUSHBUTTON, fixedFillString, 'x', NULL, NULL,
		XmVaPUSHBUTTON, suddenAppearString, 'r', NULL, NULL,
		XmVaPUSHBUTTON, diagonalString, 'g', NULL, NULL,
		XmVaPUSHBUTTON, mixedString, 'M', NULL, NULL,
		XmVaPUSHBUTTON, bonusString, 'B', NULL, NULL,
		XmVaPUSHBUTTON, cwString, 'w', NULL, NULL,
		XmVaPUSHBUTTON, plainString, 'P', NULL, NULL,
		XmVaPUSHBUTTON, showNextString, 'N', NULL, NULL,
#endif
		NULL);
	XmStringFree(pauseString);
	XmStringFree(clearString);
	XmStringFree(setupString);
#if 0
	XmStringFree(incrementString);
	XmStringFree(decrementString);
	XmStringFree(fixedFillString);
	XmStringFree(suddenAppearString);
	XmStringFree(diagonalString);
	XmStringFree(mixedString);
	XmStringFree(bonusString);
	XmStringFree(cwString);
	XmStringFree(plainString);
	XmStringFree(showNextString);
#endif
	pullDownMenu = XmCreatePulldownMenu(menuBar,
		(char *) "helpPullDown", NULL, 0);
	widget = XtVaCreateManagedWidget("Help",
		xmCascadeButtonWidgetClass, menuBar,
		XmNsubMenuId, pullDownMenu,
		XmNmnemonic, 'H', NULL);
	XtVaSetValues(menuBar, XmNmenuHelpWidget, widget, NULL);
	widget = XtVaCreateManagedWidget("Description",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'D', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpMenuListener, (char *) 0);
	widget = XtVaCreateManagedWidget("Features",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'F', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpMenuListener, (char *) 1);
	widget = XtVaCreateManagedWidget("Options",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'O', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpMenuListener, (char *) 2);
	widget = XtVaCreateManagedWidget("References",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'R', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpMenuListener, (char *) 3);
	widget = XtVaCreateManagedWidget("About",
		xmPushButtonGadgetClass, pullDownMenu,
		XmNmnemonic, 'A', NULL);
	XtAddCallback(widget, XmNactivateCallback, helpMenuListener, (char *) 4);
	XtManageChild(menuBar);
	descriptionDialog = createHelp(menuBar, (char *) descriptionHelp,
		(char *) "Description");
	featuresDialog = createHelp(menuBar, (char *) featuresHelp,
		(char *) "Features");
	optionsDialog = createHelp(menuBar, (char *) optionsHelp,
		(char *) "Options");
	referencesDialog = createHelp(menuBar, (char *) referencesHelp,
		(char *) "References");
	aboutDialog = createHelp(menuBar, (char *) aboutHelp,
		(char *) "About");
	restartDialog = createQuery(topLevel,
		(char *) "Are you sure you want to restart?",
		(char *) "Reset Query",
		(XtCallbackProc) callbackGameRestart);
	mainPanel = XtVaCreateManagedWidget("mainPanel",
		xmPanedWindowWidgetClass, menuBarPanel,
		XmNorientation, XmHORIZONTAL,
		NULL);
	controlPanel = XtVaCreateManagedWidget("controlPanel",
		xmPanedWindowWidgetClass, mainPanel,
		XmNseparatorOn, False,
		XmNsashWidth, 1,
		XmNsashHeight, 1, NULL);
	levelRowCol = XtVaCreateManagedWidget("levelRowCol",
		xmRowColumnWidgetClass, controlPanel,
		XmNnumColumns, 1,
		XmNorientation, XmHORIZONTAL,
		NULL);
	scoreRowCol = XtVaCreateManagedWidget("Rowcol",
		xmRowColumnWidgetClass, controlPanel,
		XmNnumColumns, 3,
		XmNorientation, XmHORIZONTAL,
		XmNpacking, XmPACK_COLUMN, NULL);
	levelSlider = XtVaCreateManagedWidget("levelSlider",
		xmScaleWidgetClass, levelRowCol,
		XtVaTypedArg, XmNtitleString, XmRString, "Level", 6,
		XmNminimum, MIN_LEVELS,
		XmNmaximum, MAX_LEVELS,
		XmNvalue, MIN_LEVELS,
		XmNshowValue, True,
		XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(levelSlider, XmNvalueChangedCallback,
		(XtCallbackProc) levelChangeListener, (XtPointer) NULL);
	(void) XtVaCreateManagedWidget("rowsText",
		xmLabelGadgetClass, scoreRowCol,
		XtVaTypedArg, XmNlabelString, XmRString, "Rows", 5, NULL);
	rowsText = XtVaCreateManagedWidget("0",
		xmLabelWidgetClass, scoreRowCol, NULL);
	(void) XtVaCreateManagedWidget("scoreText",
		xmLabelGadgetClass, scoreRowCol,
		XtVaTypedArg, XmNlabelString, XmRString, "Score", 6, NULL);
	scoreText = XtVaCreateManagedWidget("0",
		xmLabelWidgetClass, scoreRowCol, NULL);
	(void) XtVaCreateManagedWidget("maxScoreText",
		xmLabelGadgetClass, scoreRowCol,
		XtVaTypedArg, XmNlabelString, XmRString, "Max Score", 10, NULL);
	maxScoreText = XtVaCreateManagedWidget("0",
		xmLabelWidgetClass, scoreRowCol, NULL);
	nextWelltris = XtCreateManagedWidget("welltrisNext",
		nextWelltrisWidgetClass, controlPanel, NULL, 0);
	messageRowCol = XtVaCreateManagedWidget("messageRowCol",
		xmRowColumnWidgetClass, controlPanel, NULL);
	message = XtVaCreateManagedWidget("Play Welltris!",
		xmLabelWidgetClass, messageRowCol, NULL);
	welltris = XtCreateManagedWidget("welltris",
		welltrisWidgetClass, mainPanel, NULL, 0);
#else
	welltris = XtCreateManagedWidget("welltris",
		welltrisWidgetClass, topLevel, NULL, 0);
	shell = XtCreateApplicationShell(argv[0],
		topLevelShellWidgetClass, NULL, 0);
	nextWelltris = XtCreateManagedWidget("nextWelltris",
		nextWelltrisWidgetClass, shell, NULL, 0);
#endif
	XtVaGetValues(welltris,
#ifndef HAVE_MOTIF
		XtNshowNext, &showNext,
#endif
		XtNpixmapSize, &pixmapSize, NULL);
#ifdef HAVE_XPM
	{
		XpmAttributes xpmAttributes;
		XpmColorSymbol transparentColor[1] = {{NULL,
			(char *) "none", 0 }};
		Pixel bg;

		xpmAttributes.valuemask = XpmColorSymbols | XpmCloseness;
		xpmAttributes.colorsymbols = transparentColor;
		xpmAttributes.numsymbols = 1;
		xpmAttributes.closeness = 40000;
		XtVaGetValues(topLevel, XtNbackground, &bg, NULL);
		transparentColor[0].pixel = bg;
		(void) XpmCreatePixmapFromData(XtDisplay(topLevel),
			RootWindowOfScreen(XtScreen(topLevel)),
			RESIZE_XPM(pixmapSize), &pixmap, NULL,
			&xpmAttributes);
	}
	if (pixmap == (Pixmap) NULL)
#endif
		pixmap = XCreateBitmapFromData(XtDisplay(topLevel),
			RootWindowOfScreen(XtScreen(topLevel)),
			DEFINE_XBM);
	XtVaSetValues(topLevel,
#ifdef HAVE_MOTIF
		XmNkeyboardFocusPolicy, XmPOINTER, /* not XmEXPLICIT */
#else
		XtNinput, True,
#endif
		XtNiconPixmap, pixmap, NULL);
	XtAddCallback(welltris, XtNselectCallback,
		(XtCallbackProc) callbackGame, (XtPointer) NULL);
	XtAddCallback(nextWelltris, XtNselectCallback,
		(XtCallbackProc) callbackNextGame, (XtPointer) NULL);
	initialize(welltris);
	XtRealizeWidget(topLevel);
	VOID XGrabButton(XtDisplay(welltris), (unsigned int) AnyButton,
		AnyModifier, XtWindow(welltris), TRUE,
		(unsigned int) (ButtonPressMask),
		GrabModeAsync, GrabModeAsync, XtWindow(welltris),
		XCreateFontCursor(XtDisplay(welltris), XC_crosshair));
#ifndef HAVE_MOTIF
	XtVaGetValues(topLevel,
		XtNx, &x,
		XtNy, &y,
		XtNwidth, &width, NULL);
	XtVaSetValues(shell,
		XtNx, x + width,
		XtNy, y,
		XtNiconPixmap, pixmap,
		XtNinput, True, NULL);
	if (showNext)
		XtRealizeWidget(shell);
#endif
	(void) XtAddTimeOut((unsigned long) 1000,
		playingTimer, (XtPointer) welltris);
	XtMainLoop();

#ifdef VMS
	return 1;
#else
	return 0;
#endif
}
#endif
